import torch
import torchvision 


from visionsuite.engines.utils.torch_utils.ema import ExponentialMovingAverage
from visionsuite.engines.classification.utils.registry import MODELS
from visionsuite.engines.utils.helpers import assert_key_dict
from visionsuite.engines.utils.bases.base_oop_module import BaseOOPModule

        
@MODELS.register()
class TorchvisionModel(BaseOOPModule):
    """
    - weights can be referred by https://pytorch.org/vision/main/models.html
    """
    def __init__(self, name="TorchvisionModel"):
        super().__init__(name=name)
        self.args = None
        self._device = 'cpu'
        self._model_without_ddp = None
        self._model_ema = None
        self._model = None 
               
    def build(self, *args, **kwargs):
        super().build(*args, **kwargs)
        
        self.load_model() 
        self.to_device()
        self.set_dist()
        self.set_ema()
        
    @property
    def model(self):
        return self._model
    
    @property
    def model_without_ddp(self):
        return self._model_without_ddp
    
    @property
    def model_ema(self):
        return self._model_ema
    
    @BaseOOPModule.track_status
    def load_model(self):
        # TODO: NOT include top to specify weights to change num_classes
        try:
            if self.args['weights']:
                self._model = torchvision.models.get_model(name=self.args['model_name'] + self.args['backbone'], 
                                                       weights=self.args['weights'])
                self.log_info(f"LOADED weights: {self.args['weights']}", self.load_model.__name__, __class__.__name__)
                
                self._model.fc = torch.nn.Linear(self._model.fc.in_features, self.args['num_classes'])
                self.log_info(f"CHANGED fc for num_classes({self.args['num_classes']})", self.load_model.__name__, __class__.__name__)
            else:
                self._model = torchvision.models.get_model(name=self.args['model_name'] + self.args['backbone'], 
                                                       num_classes=self.args['num_classes'])
                
            self.log_info(f"LOADED model: {self.args['model_name']}", self.load_model.__name__, __class__.__name__)
        except Exception as error:
            raise RuntimeError(f"{error}: There has been error when loading torchvision model: {self.args['type']} with config({self.args}): ")
           
    @BaseOOPModule.track_status
    def to_device(self):
        assert_key_dict(self.args['train'], 'device')
        self._device = self.args['train']['device']
        self._model.to(self._device)

    @BaseOOPModule.track_status
    def set_dist(self):
        assert_key_dict(self.args, 'distributed')
        assert_key_dict(self.args['train'], 'sync_bn')

        if self.args['distributed'] and self.args['train']['sync_bn']:
            self._model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self._model)
            
        self._model_without_ddp = self._model
        if self.args['distributed']:
            assert_key_dict(self.args['distributed'], 'gpu')
            self._model = torch.nn.parallel.DistributedDataParallel(self._model, device_ids=[self.args['distributed']['gpu']])
            self.log_info(f"Distributed Data Parallel is set", self.set_dist.__name__, __class__.__name__)
            self._model_without_ddp = self.model.module
        else:
            assert_key_dict(self.args['train'], 'device_ids')
            if len(self.args['train']['device_ids']) > 1:
                self._model = torch.nn.parallel.DataParallel(self._model, device_ids=[self.args['train']['device_ids']])
                self._model_without_ddp = self.model.module
                self.log_info(f"Data Parallel is set", self.set_dist.__name__, __class__.__name__)
                
    @BaseOOPModule.track_status
    def set_ema(self):
        assert_key_dict(self.args, 'ema')
        self._model_ema = None
        if self.args['ema']['use']:
            assert_key_dict(self.args, 'world_size')
            assert_key_dict(self.args, 'batch_size')
            assert_key_dict(self.args, 'epochs')
                
            # Decay adjustment that aims to keep the decay independent of other hyper-parameters originally proposed at:
            # https://github.com/facebookresearch/pycls/blob/f8cd9627/pycls/core/net.py#L123
            #
            # total_ema_updates = (Dataset_size / n_GPUs) * epochs / (batch_size_per_gpu * EMA_steps)
            # We consider constant = Dataset_size for a given dataset/setup and omit it. Thus:
            # adjust = 1 / total_ema_updates ~= n_GPUs * batch_size_per_gpu * EMA_steps / epochs
            adjust = self.args['distributed']['world_size'] * self.args['train']['batch_size'] * self.args['ema_config']['steps'] / self.args['train']['epochs']
            alpha = 1.0 - self.args['ema']['decay']
            alpha = min(1.0, alpha * adjust)
            self._model_ema = ExponentialMovingAverage(self._model_without_ddp, 
                                                 device=self.args['train']['device'], 
                                                 decay=1.0 - alpha)
            
            self.log_info(f"SET EMA", self.set_ema.__name__, __class__.__name__)


from abc import ABCMeta

from omegaconf import DictConfig, OmegaConf
from pydantic import BaseModel


class BaseConfig(BaseModel, metaclass=ABCMeta):

    def add_field(self, key, value):
        self.__dict__[key] = value

    def to_dict(self):
        _dict = {}
        # for key, val in self.model_dump().items():
        #     if isinstance(val, DictConfig):
        #         _dict[key] = OmegaConf.to_container(val)
        #     else:
        #         _dict[key] = val
        for key, val in dict(self).items():
            if isinstance(val, DictConfig):
                _dict[key] = OmegaConf.to_container(val)
            elif isinstance(val, BaseModel):
                _dict[key] = val.dict()
            else:
                _dict[key] = val

        return _dict

    def json_schema(self):
        return self.model_json_schema(mode='serialization')



#!/usr/bin/env python3

"""builder.py

Currently a placeholder for subclassing `mmseg`.
"""

# I want to use the original segmentation only datasets
from mmseg.datasets import (
    DATASETS,
    PIPELINES,
    build_dataloader,
    build_dataset,
)

__all__ = [
    "DATASETS",
    "PIPELINES",
    "build_dataloader",
    "build_dataset",
]




from __future__ import absolute_import

from keras_unet_collection._model_unet_2d import unet_2d_base
from keras_unet_collection._model_vnet_2d import vnet_2d_base
from keras_unet_collection._model_unet_plus_2d import unet_plus_2d_base 
from keras_unet_collection._model_r2_unet_2d import r2_unet_2d_base
from keras_unet_collection._model_att_unet_2d import att_unet_2d_base
from keras_unet_collection._model_resunet_a_2d import resunet_a_2d_base
from keras_unet_collection._model_u2net_2d import u2net_2d_base
from keras_unet_collection._model_unet_3plus_2d import unet_3plus_2d_base
from keras_unet_collection._model_transunet_2d import transunet_2d_base
from keras_unet_collection._model_swin_unet_2d import swin_unet_2d_base



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta

import torch
from mmcv.runner import BaseModule
from mmdet.core import bbox2roi

from mmrotate.core import build_assigner, build_sampler, obb2xyxy, rbbox2result
from mmrotate.models.builder import (ROTATED_HEADS, build_head, build_roi_extractor,
                       build_shared_head)


@ROTATED_HEADS.register_module()
class CustomRotatedStandardRoIHead(BaseModule, metaclass=ABCMeta):
    """Simplest base rotated roi head including one bbox head.

    Args:
        bbox_roi_extractor (dict, optional): Config of ``bbox_roi_extractor``.
        bbox_head (dict, optional): Config of ``bbox_head``.
        shared_head (dict, optional): Config of ``shared_head``.
        train_cfg (dict, optional): Config of train.
        test_cfg (dict, optional): Config of test.
        pretrained (str, optional): Path of pretrained weight.
        init_cfg (dict, optional): Config of initialization.
        version (str, optional): Angle representations. Defaults to 'oc'.
    """

    def __init__(self,
                 bbox_roi_extractor=None,
                 bbox_head=None,
                 shared_head=None,
                 train_cfg=None,
                 test_cfg=None,
                 pretrained=None,
                 init_cfg=None,
                 version='oc'):

        super(RotatedStandardRoIHead, self).__init__(init_cfg)
        self.train_cfg = train_cfg
        self.test_cfg = test_cfg
        self.version = version

        if shared_head is not None:
            shared_head.pretrained = pretrained
            self.shared_head = build_shared_head(shared_head)

        if bbox_head is not None:
            self.init_bbox_head(bbox_roi_extractor, bbox_head)

        self.init_assigner_sampler()

        self.with_bbox = True if bbox_head is not None else False
        self.with_shared_head = True if shared_head is not None else False

    def init_assigner_sampler(self):
        """Initialize assigner and sampler."""
        self.bbox_assigner = None
        self.bbox_sampler = None
        if self.train_cfg:
            self.bbox_assigner = build_assigner(self.train_cfg.assigner)
            self.bbox_sampler = build_sampler(
                self.train_cfg.sampler, context=self)

    def init_bbox_head(self, bbox_roi_extractor, bbox_head):
        """Initialize ``bbox_head``.

        Args:
            bbox_roi_extractor (dict): Config of ``bbox_roi_extractor``.
            bbox_head (dict): Config of ``bbox_head``.
        """
        self.bbox_roi_extractor = build_roi_extractor(bbox_roi_extractor)
        self.bbox_head = build_head(bbox_head)

    def forward_dummy(self, x, proposals):
        """Dummy forward function.

        Args:
            x (list[Tensors]): list of multi-level img features.
            proposals (list[Tensors]): list of region proposals.

        Returns:
            list[Tensors]: list of region of interest.
        """
        outs = ()
        rois = bbox2roi([proposals])
        if self.with_bbox:
            bbox_results = self._bbox_forward(x, rois)
            outs = outs + (bbox_results['cls_score'],
                           bbox_results['bbox_pred'])
        return outs

    def forward_train(self,
                      x,
                      img_metas,
                      proposal_list,
                      gt_bboxes,
                      gt_labels,
                      gt_bboxes_ignore=None,
                      gt_masks=None, **kwargs):
        """
        Args:
            x (list[Tensor]): list of multi-level img features.
            img_metas (list[dict]): list of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                `mmdet/datasets/pipelines/formatting.py:Collect`.
            proposals (list[Tensors]): list of region proposals.
            gt_bboxes (list[Tensor]): Ground truth bboxes for each image with
                shape (num_gts, 5) in [cx, cy, w, h, a] format.
            gt_labels (list[Tensor]): class indices corresponding to each box
            gt_bboxes_ignore (None | list[Tensor]): specify which bounding
                boxes can be ignored when computing the loss.
            gt_masks (None | Tensor) : true segmentation masks for each box
                used if the architecture supports a segmentation task. Always
                set to None.

        Returns:
            dict[str, Tensor]: a dictionary of loss components.
        """
        # assign gts and sample proposals
        if self.with_bbox:

            num_imgs = len(img_metas)
            if gt_bboxes_ignore is None:
                gt_bboxes_ignore = [None for _ in range(num_imgs)]
            sampling_results = []
            for i in range(num_imgs):
                gt_hbboxes = obb2xyxy(gt_bboxes[i], self.version)
                assign_result = self.bbox_assigner.assign(
                    proposal_list[i], gt_hbboxes, gt_bboxes_ignore[i],
                    gt_labels[i])
                sampling_result = self.bbox_sampler.sample(
                    assign_result,
                    proposal_list[i],
                    gt_hbboxes,
                    gt_labels[i],
                    feats=[lvl_feat[i][None] for lvl_feat in x])

                if gt_bboxes[i].numel() == 0:
                    sampling_result.pos_gt_bboxes = gt_bboxes[i].new(
                        (0, gt_bboxes[0].size(-1))).zero_()
                else:
                    sampling_result.pos_gt_bboxes = \
                        gt_bboxes[i][sampling_result.pos_assigned_gt_inds, :]

                sampling_results.append(sampling_result)

        losses = dict()
        # bbox head forward and loss
        if self.with_bbox:
            bbox_results = self._bbox_forward_train(x, sampling_results,
                                                    gt_bboxes, gt_labels,
                                                    img_metas)
            losses.update(bbox_results['loss_bbox'])

        return losses

    def _bbox_forward(self, x, rois):
        """Box head forward function used in both training and testing.

        Args:
            x (list[Tensor]): list of multi-level img features.
            rois (list[Tensors]): list of region of interests.

        Returns:
            dict[str, Tensor]: a dictionary of bbox_results.
        """
        bbox_feats = self.bbox_roi_extractor(
            x[:self.bbox_roi_extractor.num_inputs], rois)
        if self.with_shared_head:
            bbox_feats = self.shared_head(bbox_feats)
        cls_score, bbox_pred = self.bbox_head(bbox_feats)

        bbox_results = dict(
            cls_score=cls_score, bbox_pred=bbox_pred, bbox_feats=bbox_feats)
        return bbox_results

    def _bbox_forward_train(self, x, sampling_results, gt_bboxes, gt_labels,
                            img_metas):
        """Run forward function and calculate loss for box head in training.

        Args:
            x (list[Tensor]): list of multi-level img features.
            sampling_results (list[Tensor]): list of sampling results.
            gt_bboxes (list[Tensor]): Ground truth bboxes for each image with
                shape (num_gts, 5) in [cx, cy, w, h, a] format.
            gt_labels (list[Tensor]): class indices corresponding to each box
            img_metas (list[dict]): list of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.

        Returns:
            dict[str, Tensor]: a dictionary of bbox_results.
        """
        rois = bbox2roi([res.bboxes for res in sampling_results])
        bbox_results = self._bbox_forward(x, rois)

        bbox_targets = self.bbox_head.get_targets(sampling_results, gt_bboxes,
                                                  gt_labels, self.train_cfg)
        loss_bbox = self.bbox_head.loss(bbox_results['cls_score'],
                                        bbox_results['bbox_pred'], rois,
                                        *bbox_targets)

        bbox_results.update(loss_bbox=loss_bbox)
        return bbox_results

    async def async_simple_test(self,
                                x,
                                proposal_list,
                                img_metas,
                                rescale=False):
        """Async test without augmentation.

        Args:
            x (list[Tensor]): list of multi-level img features.
            proposal_list (list[Tensors]): list of region proposals.
            img_metas (list[dict]): list of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
            rescale (bool): If True, return boxes in original image space.
                Default: False.

        Returns:
            dict[str, Tensor]: a dictionary of bbox_results.
        """
        assert self.with_bbox, 'Bbox head must be implemented.'

        det_bboxes, det_labels = await self.async_test_bboxes(
            x, img_metas, proposal_list, self.test_cfg, rescale=rescale)
        bbox_results = rbbox2result(det_bboxes, det_labels,
                                    self.bbox_head.num_classes)
        return bbox_results

    def simple_test(self, x, proposal_list, img_metas, rescale=False):
        """Test without augmentation.

        Args:
            x (list[Tensor]): list of multi-level img features.
            proposal_list (list[Tensors]): list of region proposals.
            img_metas (list[dict]): list of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
            rescale (bool): If True, return boxes in original image space.
                Default: False.

        Returns:
            dict[str, Tensor]: a dictionary of bbox_results.
        """
        assert self.with_bbox, 'Bbox head must be implemented.'

        det_bboxes, det_labels = self.simple_test_bboxes(
            x, img_metas, proposal_list, self.test_cfg, rescale=rescale)

        bbox_results = [
            rbbox2result(det_bboxes[i], det_labels[i],
                         self.bbox_head.num_classes)
            for i in range(len(det_bboxes))
        ]

        return bbox_results

    def aug_test(self, x, proposal_list, img_metas, rescale=False):
        """Test with augmentations."""
        raise NotImplementedError

    def simple_test_bboxes(self,
                           x,
                           img_metas,
                           proposals,
                           rcnn_test_cfg,
                           rescale=False):
        """Test only det bboxes without augmentation.

        Args:
            x (tuple[Tensor]): Feature maps of all scale level.
            img_metas (list[dict]): Image meta info.
            proposals (List[Tensor]): Region proposals.
            rcnn_test_cfg (obj:`ConfigDict`): `test_cfg` of R-CNN.
            rescale (bool): If True, return boxes in original image space.
                Default: False.

        Returns:
            tuple[list[Tensor], list[Tensor]]: The first list contains \
                the boxes of the corresponding image in a batch, each \
                tensor has the shape (num_boxes, 5) and last dimension \
                5 represent (tl_x, tl_y, br_x, br_y, score). Each Tensor \
                in the second list is the labels with shape (num_boxes, ). \
                The length of both lists should be equal to batch_size.
        """

        rois = bbox2roi(proposals)

        if rois.shape[0] == 0:
            batch_size = len(proposals)
            det_bbox = rois.new_zeros(0, 5)
            det_label = rois.new_zeros((0, ), dtype=torch.long)
            if rcnn_test_cfg is None:
                det_bbox = det_bbox[:, :4]
                det_label = rois.new_zeros(
                    (0, self.bbox_head.fc_cls.out_features))
            # There is no proposal in the whole batch
            return [det_bbox] * batch_size, [det_label] * batch_size

        bbox_results = self._bbox_forward(x, rois)
        img_shapes = tuple(meta['img_shape'] for meta in img_metas)
        scale_factors = tuple(meta['scale_factor'] for meta in img_metas)

        # split batch bbox prediction back to each image
        cls_score = bbox_results['cls_score']
        bbox_pred = bbox_results['bbox_pred']
        num_proposals_per_img = tuple(len(p) for p in proposals)
        rois = rois.split(num_proposals_per_img, 0)
        cls_score = cls_score.split(num_proposals_per_img, 0)

        # some detector with_reg is False, bbox_pred will be None
        if bbox_pred is not None:
            # TODO move this to a sabl_roi_head
            # the bbox prediction of some detectors like SABL is not Tensor
            if isinstance(bbox_pred, torch.Tensor):
                bbox_pred = bbox_pred.split(num_proposals_per_img, 0)
            else:
                bbox_pred = self.bbox_head.bbox_pred_split(
                    bbox_pred, num_proposals_per_img)
        else:
            bbox_pred = (None, ) * len(proposals)

        # apply bbox post-processing to each image individually
        det_bboxes = []
        det_labels = []
        for i in range(len(proposals)):
            if rois[i].shape[0] == 0:
                # There is no proposal in the single image
                det_bbox = rois[i].new_zeros(0, 5)
                det_label = rois[i].new_zeros((0, ), dtype=torch.long)
                if rcnn_test_cfg is None:
                    det_bbox = det_bbox[:, :4]
                    det_label = rois[i].new_zeros(
                        (0, self.bbox_head.fc_cls.out_features))

            else:
                det_bbox, det_label = self.bbox_head.get_bboxes(
                    rois[i],
                    cls_score[i],
                    bbox_pred[i],
                    img_shapes[i],
                    scale_factors[i],
                    rescale=rescale,
                    cfg=rcnn_test_cfg)
            det_bboxes.append(det_bbox)
            det_labels.append(det_label)
        return det_bboxes, det_labels



import numpy as np
from collections import Counter
import cv2
import os.path as osp


# https://github.com/rafaelpadilla/review_object_detection_metrics?tab=readme-ov-file#ap-with-iou-threshold-t05

def is_overlapped(box_1, box_2):
    if box_1[0] > box_2[2]:
        return False
    if box_2[0] > box_1[2]:
        return False
    if box_1[3] < box_2[1]:
        return False
    if box_1[1] > box_2[3]:
        return False
    
    return True

def get_area(box):
    
    area = (box[2] - box[0])*(box[3] - box[1])
    
    return area 

def get_overlap_area(box_1, box_2):
    lt_x = max(box_1[0], box_2[0])
    lt_y = max(box_1[1], box_2[1])
    rb_x = min(box_1[2], box_2[2])
    rb_y = min(box_1[3], box_2[3])

    overlap_area = (rb_x - lt_x)*(rb_y - lt_y)
    
    return overlap_area

def get_coord_diff(box_1, box_2):
    import numpy as np
    from scipy.spatial import distance

    box_1 = [box_1[i:i + 2] for i in range(0, len(box_1), 2)]
    box_2 = [box_2[i:i + 2] for i in range(0, len(box_2), 2)]

    polygon1 = np.array(box_1)
    polygon2 = np.array(box_2)

    def average_difference(poly1, poly2):
        total_diff = 0
        matched_poly2 = []

        for p1 in poly1:
            distances = distance.cdist([p1], poly2, 'euclidean')
            nearest_idx = np.argmin(distances)
            matched_poly2.append(poly2[nearest_idx])
            total_diff += distances[0, nearest_idx]

        average_diff = total_diff / len(poly1)  # 평균 차이 계산
        return average_diff

    return average_difference(polygon1, polygon2)


def get_iou(box_1, box_2, shape_type, return_dict=False):
    
    if shape_type == 'rectangle':
        if not is_overlapped(box_1, box_2):
            return 0
        
        area_1 = get_area(box_1)
        area_2 = get_area(box_2)
        
        overlap_area = get_overlap_area(box_1, box_2)
        assert overlap_area > 0, RuntimeError(f"ERROR: overlap-area must be more than 0, not {overlap_area}")
        
        
        iou = overlap_area/float(area_1 + area_2 - overlap_area)
        assert iou >= 0, RuntimeError(f"ERROR: iou must be more than 0, not {iou}")
        
        if return_dict:
            return {'iou': iou, 'area_1': area_1, 'area_2': area_2, 'overlap_area': overlap_area}
        else:
            return iou
    elif shape_type == 'polygon':
        iou, area_1, area_2, overlap_area = get_polygon_iou(box_1, box_2)
        
        if return_dict:
            return {'iou': iou, 'area_1': area_1, 'area_2': area_2, 'overlap_area': overlap_area}
        else:
            return iou        

def ond_dim_points_to_polygon(points):
    # points를 (x, y) 형식의 튜플 리스트로 변환
    return [(points[i], points[i + 1]) for i in range(0, len(points), 2)]


def get_polygon_iou(point1, point2):
    from shapely.geometry import Polygon
    # 두 다각형을 Polygon 객체로 변환
    poly1 = Polygon(ond_dim_points_to_polygon(point1))
    poly2 = Polygon(ond_dim_points_to_polygon(point2))
    
    if not poly1.is_valid:
        poly1 = poly1.buffer(0)
        
    if not poly2.is_valid:
        poly2 = poly2.buffer(0)
        
    # 교집합 영역 계산
    intersection_area = poly1.intersection(poly2).area
    
    
    union_area = poly1.union(poly2).area
    
    # IoU 계산
    iou = intersection_area / union_area
    
    return iou, poly1.area,poly2.area, intersection_area
    
def ElevenPointInterpolatedAP(rec, prec):
    mrec = [e for e in rec]
    mpre = [e for e in prec]

    # recallValues = [1.0, 0.9, ..., 0.0]
    recallValues = np.linspace(0, 1, 11)
    recallValues = list(recallValues[::-1])
    rhoInterp, recallValid = [], []


    for r in recallValues:
        # r : recall값의 구간
        # argGreaterRecalls : r보다 큰 값의 index
        argGreaterRecalls = np.argwhere(mrec[:] >= r)
        pmax = 0
        print(r, argGreaterRecalls)

        # precision 값 중에서 r 구간의 recall 값에 해당하는 최댓값
        if argGreaterRecalls.size != 0:
            pmax = max(mpre[argGreaterRecalls.min():])

        recallValid.append(r)
        rhoInterp.append(pmax)

    ap = sum(rhoInterp) / 11
    
    return [ap, rhoInterp, recallValues, None]

def mAP(result):
    ap = 0
    for r in result:
        ap += r['ap']
    mAP = ap / len(result)
    
    return mAP

def get_average_g_results(g_results):
    result = {}
    total_avg_g_fnr, total_avg_g_fpr = 0, 0
    cnt = 0
    for image_name, g_result in g_results.items():
        avg_g_fnr, avg_g_fpr = 0, 0
        for key, val in g_result.items():
            avg_g_fnr += val['g_fnr']
            avg_g_fpr += val['g_fpr']
            cnt += 1
        total_avg_g_fnr += avg_g_fnr
        total_avg_g_fpr += avg_g_fpr
        result[image_name] = {'avg_g_fnr': avg_g_fnr/len(val), 
                              'avg_g_fpr': avg_g_fpr/len(val)
                             }
        
    result['total_avg_g_fnr'] = total_avg_g_fnr/cnt
    result['total_avg_g_fpr'] = total_avg_g_fpr/cnt
    
    return result
        
        
def update_ap_by_image(results_by_image):
    
    overall_by_class = {}
    if len(results_by_image) != 0:
        for image_name, val in results_by_image.items():
            for _class, results in val.items():
                fpr = results['fp']/(results['tp'] + results['fp'] + 1e-5)
                fnr = results['fn']/(results['tp'] + results['fn'] + 1e-5)
                results.update({'fpr': fpr, 'fnr': fnr})
                
                if _class not in overall_by_class:
                    overall_by_class[_class] = {'fpr': [], 'fnr': [], 'tp': [], 'fp': [], 'fn': [], 'tn': [], 'total_gt': [], 
                                                'miou': [], 'mean_coord_diff': []}
                    
                overall_by_class[_class]['fpr'].append(fpr)
                overall_by_class[_class]['fnr'].append(fnr)
                overall_by_class[_class]['tp'].append(results['tp'])
                overall_by_class[_class]['fp'].append(results['fp'])
                overall_by_class[_class]['fn'].append(results['fn'])
                overall_by_class[_class]['tn'].append(results['tn'])
                overall_by_class[_class]['total_gt'].append(results['total_gt'])
                overall_by_class[_class]['miou'].append(results['miou'])
                overall_by_class[_class]['mean_coord_diff'].append(results['mean_coord_diff'])
                
        for key, val in overall_by_class.items():
            overall_by_class[key]['fpr'] = np.mean(overall_by_class[key]['fpr'])
            overall_by_class[key]['fnr'] = np.mean(overall_by_class[key]['fnr'])
            overall_by_class[key]['tp'] = np.sum(overall_by_class[key]['tp'])
            overall_by_class[key]['fp'] = np.sum(overall_by_class[key]['fp'])
            overall_by_class[key]['fn'] = np.sum(overall_by_class[key]['fn'])
            overall_by_class[key]['tn'] = np.sum(overall_by_class[key]['tn'])
            overall_by_class[key]['total_gt'] = np.sum(overall_by_class[key]['total_gt'])
            overall_by_class[key]['stdiou'] = np.std(overall_by_class[key]['miou'])
            overall_by_class[key]['miou'] = np.mean(overall_by_class[key]['miou'])
            overall_by_class[key]['std_coord_diff'] = np.std(overall_by_class[key]['mean_coord_diff'])
            overall_by_class[key]['mean_coord_diff'] = np.mean(overall_by_class[key]['mean_coord_diff'])
                
    results_by_image['overall'] = overall_by_class
    
    return results_by_image
    _

def update_ious_by_image(results_by_image):
    
    if len(results_by_image) != 0:
        for image_name, val in results_by_image.items():
            for _class, results in val.items():
                results.update({'miou': np.mean(results['iou']) if len(results['iou']) != 0 else 0})
                    
    return results_by_image    

def update_coord_diff_by_image(results_by_image):
    
    if len(results_by_image) != 0:
        for image_name, val in results_by_image.items():
            for _class, results in val.items():
                results.update({'mean_coord_diff': np.mean(results['coord_diff'])})
                    
    return results_by_image  
    
def calculateAveragePrecision(rec, prec):
    
    mrec = [0] + [e for e in rec] + [1]
    mpre = [0] + [e for e in prec] + [0]

    for i in range(len(mpre)-1, 0, -1):
        mpre[i-1] = max(mpre[i-1], mpre[i])

    ii = []

    for i in range(len(mrec)-1):
        if mrec[1:][i] != mrec[0:-1][i]:
            ii.append(i+1)

    ap = 0
    for i in ii:
        ap = ap + np.sum((mrec[i] - mrec[i-1]) * mpre[i])
    
    return [ap, mpre[0:len(mpre)-1], mrec[0:len(mpre)-1], ii]

def get_performance(detections, ground_truths, classes, iou_threshold=0.3, method='ap', shape_type='rectangle'):
    '''
        detections: ['image filename', class-index, confidence, (x1, y1, x2, y2)]
        ground_truths: ['image filename', class-index, confidence, (x1, y1, x2, y2)]
    '''
    
    results_by_class = []
    results_by_image = {}
    # loop by each class for all images
    for _class in classes:
        
        # Extract info for each class
        dets = [det for det in detections if det[1] == _class]
        gts = [gt for gt in ground_truths if gt[1] == _class]
        
        num_gt = len(gts) # len(tp) + len(tn)
        
        # dets = sorted(dets, key=lambda conf: conf[2], reverse=True) # descending by confidence
        if all(conf[2] is not None for conf in dets):  
            dets = sorted(dets, key=lambda conf: conf[2], reverse=True) 

        tp, fp = np.zeros(len(dets)), np.zeros(len(dets))
        gt_box_detected_map = Counter(c[0] for c in gts) # number of gt-boxes by image
        for key, val in gt_box_detected_map.items():
            gt_box_detected_map[key] = np.zeros(val)
            
        for det_index, det in enumerate(dets):
            
            # match dets and gts by image
            gt = [gt for gt in gts if gt[0] == det[0]]
            
            if det[0] not in results_by_image:
                results_by_image[det[0]] = {_class: {'tp': 0, 'fp': 0, 'fn': 0, 'tn': 0, 'total_gt': len(gt), 'iou': [], 'coord_diff': []}}
            
            else: 
                if _class not in results_by_image[det[0]]:
                    results_by_image[det[0]].update({_class: {'tp': 0, 'fp': 0, 'fn': 0, 'tn': 0, 'total_gt': len(gt), 'iou': [], 'coord_diff': []}})

            max_iou, iou = 0, 0
            max_gt_coord, max_pred_coord = None, None
            for gt_index, _gt in enumerate(gt):
                '''
                    Within the same image, compare all gt-boxes for each det-box and then, calculate iou.
                    match the det-box and gt-box by the maximum iou.
                '''
                iou = get_iou(det[3], _gt[3], shape_type)
                if iou > max_iou:
                    max_iou = iou 
                    max_gt_index = gt_index
                    max_gt_coord, max_pred_coord = det[3], _gt[3]
                    
                # FIXME: move into the max_iou >= iou_threshold phrase???
                if iou >= iou_threshold:
                    results_by_image[det[0]][_class]['tp'] += 1
                
            if results_by_image[det[0]][_class]['tp'] == 0:
                results_by_image[det[0]][_class]['fp'] += 1
                
            if max_iou >= iou_threshold:
                '''
                    * tp: 
                        - bigger than iou-threhold 
                        - gt-box is not matched yet
                    * fp:
                        - otherwise
                '''
                # iou
                results_by_image[det[0]][_class]['iou'].append(max_iou)
                
                # coord-diff
                if max_gt_coord is not None and  max_pred_coord is not None:
                    results_by_image[det[0]][_class]['coord_diff'].append(get_coord_diff(max_gt_coord, max_pred_coord))
                
                if gt_box_detected_map[det[0]][max_gt_index] == 0:
                    tp[det_index] = 1
                    gt_box_detected_map[det[0]][max_gt_index] = 1
                else:
                    fp[det_index] = 1
            else:
                fp[det_index] = 1
                
        if len(gt_box_detected_map) != 0:
            if isinstance(gt_box_detected_map[det[0]] , int) and gt_box_detected_map[det[0]] != 0:
                results_by_image[det[0]][_class]['fn'] = len(gt_box_detected_map[det[0]]) - np.sum(gt_box_detected_map[det[0]])
            elif isinstance(gt_box_detected_map[det[0]] , np.ndarray):
                results_by_image[det[0]][_class]['fn'] = len(gt_box_detected_map[det[0]]) - np.sum(gt_box_detected_map[det[0]])
                    
        accumulated_tp = np.cumsum(tp)
        accumulated_fp = np.cumsum(fp)
        accumulated_precision = np.divide(accumulated_tp, (accumulated_tp + accumulated_fp))
        accumulated_recall = accumulated_tp/num_gt if num_gt != 0 else accumulated_tp
                    
        if method.lower() == 'ap':
            [ap, mean_precision, mean_recall, ii] = calculateAveragePrecision(accumulated_recall, accumulated_precision)
        else:
            [ap, mean_precision, mean_recall, _] = ElevenPointInterpolatedAP(accumulated_recall, accumulated_precision)

        result_by_class = {
            'class' : _class,
            'accumulated_precision' : accumulated_precision,
            'accumulated_recall' : accumulated_recall,
            'precision': accumulated_precision[-1] if len(accumulated_precision) != 0 else 0,
            'recall': accumulated_recall[-1] if len(accumulated_recall) != 0 else 0,
            'ap' : ap,
            'interpolated_precision' : mean_precision,
            'interpolated_recall' : mean_recall,
            'total_gt' : num_gt,
            'total_tp' : np.sum(tp),
            'total_fp' : np.sum(fp),
            'total_fn' : num_gt - np.sum(tp),
        }
        
        results_by_class.append(result_by_class)


    for image_name, results in results_by_image.items():
        for _class in classes:
            gts = [gt for gt in ground_truths if gt[1] == _class]
            gt = [gt for gt in gts if gt[0] == image_name]

            if _class not in results:
                results.update({_class: {'tp': 0, 'fp': 0, 'fn': len(gt), 'tn': 0, 'total_gt': len(gt), 'iou': [], 'coord_diff': []}})
        

    results_by_image = update_ious_by_image(results_by_image)
    results_by_image = update_coord_diff_by_image(results_by_image)
    results_by_image = update_ap_by_image(results_by_image)
    results_by_class.append({'map': mAP(results_by_class)})
    
        
    
        
    return {'by_class': results_by_class, 'by_image': dict(sorted(results_by_image.items()))}






from visionsuite.engines.segmentation.utils.registry import DATASETS
from visionsuite.engines.utils.helpers import get_params_from_obj

def build_dataset(transform=None, **config):
    dataset_obj = DATASETS.get(config['type'], case_sensitive=config['case_sensitive'])
    dataset_params = get_params_from_obj(dataset_obj)
    for key in dataset_params.keys():
        if key in config:
            dataset_params[key] = config[key]
            
        if key == 'transform':
            dataset_params[key] = transform
    
    dataset = dataset_obj(**dataset_params)
    
    return dataset
    
    
    
# from visionsuite.engines.segmentation.src.datasets.coco_utils import get_coco
# from visionsuite.engines.segmentation.src.datasets.mask_dataset import get_mask
# import torchvision

# def get_dataset(args, is_train):
#     def sbd(*args, **kwargs):
#         kwargs.pop("use_v2")
#         return torchvision.datasets.SBDataset(*args, mode="segmentation", **kwargs)

#     def voc(*args, **kwargs):
#         kwargs.pop("use_v2")
#         return torchvision.datasets.VOCSegmentation(*args, **kwargs)

#     paths = {
#         "voc": (args['input_dir'], voc, 21),
#         "voc_aug": (args['input_dir'], sbd, 21),
#         "coco": (args['input_dir'], get_coco, 21),
#         "mask": (args['input_dir'], get_mask, 4),
#     }
#     p, ds_fn, num_classes = paths[args['dataset']]

#     image_set = "train" if is_train else "val"
#     ds = ds_fn(p, image_set=image_set, transforms=get_transform(is_train, args), use_v2=args['use_v2'])
#     return ds, num_classes




import inspect
import threading
import warnings
from collections import OrderedDict
from typing import Type, TypeVar

_lock = threading.RLock()
T = TypeVar("T")


def _accquire_lock() -> None:
    """
    Acquire the module-level lock for serializing access to shared data.

    This should be released with _release_lock().
    """
    if _lock:
        _lock.acquire()


def _release_lock() -> None:
    """Release the module-level lock acquired by calling _accquire_lock()."""
    if _lock:
        _lock.release()


class ManagerMeta(type):
    """
    The metaclass for global accessible class.

    The subclasses inheriting from ``ManagerMeta`` will manage their
    own ``_instance_dict`` and root instances. The constructors of subclasses
    must contain the ``name`` argument.

    Examples:
        >>> class SubClass1(metaclass=ManagerMeta):
        >>>     def __init__(self, *args, **kwargs):
        >>>         pass
        AssertionError: <class '__main__.SubClass1'>.__init__ must have the
        name argument.
        >>> class SubClass2(metaclass=ManagerMeta):
        >>>     def __init__(self, name):
        >>>         pass
        >>> # valid format.
    """

    def __init__(cls, *args):
        cls._instance_dict = OrderedDict()
        params = inspect.getfullargspec(cls)
        params_names = params[0] if params[0] else []
        assert "name" in params_names, f"{cls} must have the `name` argument"
        super().__init__(*args)


class ManagerMixin(metaclass=ManagerMeta):
    """
    ``ManagerMixin`` is the base class for classes that have global access requirements.

    The subclasses inheriting from ``ManagerMixin`` can get their
    global instances.

    Examples:
        >>> class GlobalAccessible(ManagerMixin):
        >>>     def __init__(self, name=''):
        >>>         super().__init__(name)
        >>>
        >>> GlobalAccessible.get_instance('name')
        >>> instance_1 = GlobalAccessible.get_instance('name')
        >>> instance_2 = GlobalAccessible.get_instance('name')
        >>> assert id(instance_1) == id(instance_2)

    Args:
        name (str): Name of the instance. Defaults to ''.
    """

    def __init__(self, name: str = "", **kwargs):
        assert (
            isinstance(name, str) and name
        ), "name argument must be an non-empty string."
        self._instance_name = name

    @classmethod
    def get_instance(cls: Type[T], name: str, **kwargs) -> T:
        """
        Get subclass instance by name if the name exists.

        If corresponding name instance has not been created, ``get_instance``
        will create an instance, otherwise ``get_instance`` will return the
        corresponding instance.

        Examples
            >>> instance1 = GlobalAccessible.get_instance('name1')
            >>> # Create name1 instance.
            >>> instance.instance_name
            name1
            >>> instance2 = GlobalAccessible.get_instance('name1')
            >>> # Get name1 instance.
            >>> assert id(instance1) == id(instance2)

        Args:
            name (str): Name of instance. Defaults to ''.

        Returns:
            object: Corresponding name instance, the latest instance, or root
            instance.
        """
        _accquire_lock()
        assert isinstance(name, str), f"type of name should be str, but got {type(cls)}"
        instance_dict = cls._instance_dict  # type: ignore
        # Get the instance by name.
        if name not in instance_dict:
            instance = cls(name=name, **kwargs)  # type: ignore
            instance_dict[name] = instance  # type: ignore
        elif kwargs:
            warnings.warn(
                f"{cls} instance named of {name} has been created, "
                "the method `get_instance` should not accept any other "
                "arguments"
            )
        # Get latest instantiated instance or root instance.
        _release_lock()
        return instance_dict[name]

    @classmethod
    def get_current_instance(cls):
        """
        Get latest created instance.

        Before calling ``get_current_instance``, The subclass must have called
        ``get_instance(xxx)`` at least once.

        Examples
            >>> instance = GlobalAccessible.get_current_instance()
            AssertionError: At least one of name and current needs to be set
            >>> instance = GlobalAccessible.get_instance('name1')
            >>> instance.instance_name
            name1
            >>> instance = GlobalAccessible.get_current_instance()
            >>> instance.instance_name
            name1

        Returns:
            object: Latest created instance.
        """
        _accquire_lock()
        if not cls._instance_dict:
            raise RuntimeError(
                f"Before calling {cls.__name__}.get_current_instance(), you "
                "should call get_instance(name=xxx) at least once."
            )
        name = next(iter(reversed(cls._instance_dict)))
        _release_lock()
        return cls._instance_dict[name]

    @classmethod
    def check_instance_created(cls, name: str) -> bool:
        """
        Check whether the name corresponding instance exists.

        Args:
            name (str): Name of instance.

        Returns:
            bool: Whether the name corresponding instance exists.
        """
        return name in cls._instance_dict

    @property
    def instance_name(self) -> str:
        """
        Get the name of instance.

        Returns:
            str: Name of instance.
        """
        return self._instance_name


import os.path as osp
import glob 
import json
import warnings

def labelme2metrics(input_dir, return_class2idx=True):
    json_files = glob.glob(osp.join(input_dir, '*.json'))

    gts = []
    class2idx = {}
    num_labelme_label = 0
    for json_file in json_files:
        
        with open(json_file, 'r') as jf:
            anns = json.load(jf)['shapes']
            
        if len(anns) != 0:
            filename = osp.split(osp.splitext(json_file)[0])[-1]
            for ann in anns:
                shape_type = ann['shape_type']
                points = ann['points']
                label = ann['label']
                num_labelme_label += 1
                
                if label not in class2idx:
                    class2idx[label] = len(class2idx)
                
                if shape_type == 'rectangle':
                    assert len(points) == 2 and len(points[0]) == 2, ValueError(f"ERROR: the number of points must be 2, not {len(points)}: {points} at {json_file}")
                    
                    gt = [filename, class2idx[label], 1, (min(points[0][0], points[1][0]),
                                                          min(points[0][1], points[1][1]), 
                                                          max(points[0][0], points[1][0]),
                                                          max(points[0][1], points[1][1]))
                        ]
                    gts.append(gt)          
                elif shape_type == 'polygon':
                    assert len(points) >=3, ValueError(f"ERROR: the number of points must be more than 3, not {len(points)}: {points} at {json_file}")
                    
                    gt = [filename, class2idx[label], 1, tuple([_point for __point in points for _point in __point])]
                    gts.append(gt)
                else:
                    NotImplementedError(f"ERROR: NOT Consider {shape_type} shape-type at {json_file}")
                    
        if len(gts) != num_labelme_label:
            warnings.warn(f"The number of lableme label({num_labelme_label}) is not same to the number of metrics label({len(gts)})")    
                              

    if return_class2idx:
        return gts, class2idx
    else:
        return gts
        
if __name__ == '__main__':
    input_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/w_json/학습'


    ground_truths = labelme2metrics(input_dir)
    print(ground_truths)



# Copyright (c) ByteDance, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""
Mostly copy-paste from DINO and timm library:
https://github.com/facebookresearch/dino
https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py
"""


import math
import torch
import torch.nn as nn

from functools import partial
from timm.models.vision_transformer import trunc_normal_
from timm.models.registry import register_model

def drop_path(x, drop_prob: float = 0., training: bool = False):
    if drop_prob == 0. or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """
    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]

        attn = ((self.scale * q) @ k.transpose(-2, -1))
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x, attn

class Block(nn.Module):
    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., 
                 attn_drop=0., drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, init_values=0):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if init_values > 0:
            self.gamma_1 = nn.Parameter(init_values * torch.ones((dim)), requires_grad=True)
            self.gamma_2 = nn.Parameter(init_values * torch.ones((dim)), requires_grad=True)
        else:
            self.gamma_1, self.gamma_2 = None, None

    def forward(self, x, return_attention=False):
        y, attn = self.attn(self.norm1(x))
        if return_attention:
            return attn
        if self.gamma_1 is None:
            x = x + self.drop_path(y)
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        else:
            x = x + self.drop_path(self.gamma_1 * y)
            x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))
        return x

class PatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        num_patches = (img_size // patch_size) * (img_size // patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)
            
    def forward(self, x):
        B, C, H, W = x.shape
        return self.proj(x)

class VisionTransformer(nn.Module):
    """ Vision Transformer """
    def __init__(self, img_size=[224], patch_size=16, in_chans=3, num_classes=0, embed_dim=768, depth=12,
                 num_heads=12, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop_rate=0., attn_drop_rate=0.,
                 drop_path_rate=0., norm_layer=partial(nn.LayerNorm, eps=1e-6), return_all_tokens=False, 
                 init_values=0, use_mean_pooling=False):
        super().__init__()
        self.num_features = self.embed_dim = embed_dim
        self.return_all_tokens = return_all_tokens

        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim)
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_rate)

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = nn.ModuleList([
            Block(
                dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer, 
                init_values=init_values)
            for i in range(depth)])

        self.norm = nn.Identity() if use_mean_pooling else norm_layer(embed_dim)
        self.fc_norm = norm_layer(embed_dim) if use_mean_pooling else None
        # Classifier head
        self.head = nn.Linear(embed_dim, num_classes) if num_classes > 0 else nn.Identity()

        trunc_normal_(self.pos_embed, std=.02)
        trunc_normal_(self.cls_token, std=.02)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def interpolate_pos_encoding(self, x, w, h):
        npatch = x.shape[1] - 1
        N = self.pos_embed.shape[1] - 1
        if npatch == N and w == h:
            return self.pos_embed
        class_pos_embed = self.pos_embed[:, 0]
        patch_pos_embed = self.pos_embed[:, 1:]
        dim = x.shape[-1]
        w0 = w // self.patch_embed.patch_size
        h0 = h // self.patch_embed.patch_size
        # we add a small number to avoid floating point error in the interpolation
        # see discussion at https://github.com/facebookresearch/dino/issues/8
        w0, h0 = w0 + 0.1, h0 + 0.1
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
            scale_factor=(w0 / math.sqrt(N), h0 / math.sqrt(N)),
            mode='bicubic',)
        assert int(w0) == patch_pos_embed.shape[-2] and int(h0) == patch_pos_embed.shape[-1]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1)

    def prepare_tokens(self, x, mask=None):
        B, nc, w, h = x.shape
        # patch linear embedding
        x = self.patch_embed(x)

        # mask image modeling
        if mask is not None:
            x = self.mask_model(x, mask)
        x = x.flatten(2).transpose(1, 2)

        # add the [CLS] token to the embed patch tokens
        cls_tokens = self.cls_token.expand(B, -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        # add positional encoding to each token
        x = x + self.interpolate_pos_encoding(x, w, h)

        return self.pos_drop(x)

    def forward(self, x, return_all_tokens=None):
        x = self.prepare_tokens(x)

        for blk in self.blocks:
            x = blk(x)

        x = self.norm(x)
        if self.fc_norm is not None:
            x[:, 0] = self.fc_norm(x[:, 1:, :].mean(1))
        
        return_all_tokens = self.return_all_tokens if \
            return_all_tokens is None else return_all_tokens
        if return_all_tokens:
            return x
        return x[:, 0]

    def get_last_selfattention(self, x):
        x = self.prepare_tokens(x)
        for i, blk in enumerate(self.blocks):
            if i < len(self.blocks) - 1:
                x = blk(x)
            else:
                # return attention of the last block
                return blk(x, return_attention=True)

    def get_intermediate_layers(self, x, n=1):
        x = self.prepare_tokens(x)
        # we return the output tokens from the `n` last blocks
        output = []
        for i, blk in enumerate(self.blocks):
            x = blk(x)
            if len(self.blocks) - i <= n:
                output.append(self.norm(x))
        return output
        
    def get_num_layers(self):
        return len(self.blocks)

def vit_tiny(patch_size=16, **kwargs):
    model = VisionTransformer(
        patch_size=patch_size, embed_dim=192, depth=12, num_heads=3, mlp_ratio=4,
        qkv_bias=True, **kwargs)
    return model

def vit_small(patch_size=16, **kwargs):
    model = VisionTransformer(
        patch_size=patch_size, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4,
        qkv_bias=True, **kwargs)
    return model

def vit_base(patch_size=16, **kwargs):
    model = VisionTransformer(
        patch_size=patch_size, embed_dim=768, depth=12, num_heads=12, mlp_ratio=4,
        qkv_bias=True, **kwargs)
    return model

def vit_large(patch_size=16, **kwargs):
    model = VisionTransformer(
        patch_size=patch_size, embed_dim=1024, depth=24, num_heads=16, mlp_ratio=4,
        qkv_bias=True, **kwargs)
    return model



from .labelme2dota import convert_labelme2dota
from .dota2yolo import convert_dota2yolo_obb

def convert_labelme2yolo_obb(input_dir: str, output_dir: str=None):

    output_dir = input_dir + '_dota'

    convert_labelme2dota(input_dir, output_dir, copy_image=True, image_ext='bmp')

    input_dir = output_dir
    output_dir = input_dir + '_yolo_obb'

    convert_dota2yolo_obb(input_dir, output_dir, copy_image=True, image_ext='bmp')
    
if __name__ == '__main__':
    input_dir = '/HDD/datasets/projects/kt_g/datasets/split_dataset'
    copy_image = True
    image_ext = 'bmp'
    
    convert_labelme2yolo_obb(input_dir, copy_image=copy_image, image_ext=image_ext)



# import argparse
# import yaml
# import numpy as np

# from athena.src.data import set_roi_params
# from athena.src.data import set_patch_params
# from athena.src.tasks.segmentation.utils.params.db import set_db_params

# from athena.src.data import get_images_info

# class PatchInfo:
#     def __init__(self):
#         self.cfgs = argparse.Namespace()
#         self._vars = argparse.Namespace()

#     def alg_set_cfgs(self, recipe):

#         _cfgs = vars(self.cfgs)
#         if recipe is not None:
#             if isinstance(recipe, str):
#                 with open(recipe) as f:
#                     recipe = yaml.safe_load(f)

#             for key, val in recipe.items():
#                 _cfgs[key] = val

#         set_roi_params(self.cfgs, self._vars)
#         set_patch_params(self.cfgs, self._vars)
#         set_db_params(self.cfgs, self._vars)

#         if hasattr(self.cfgs, "input_dir"):
#             setattr(self._vars, "img_folder", str(getattr(self.cfgs,  "input_dir")))
#         else:
#             setattr(self._vars, "img_folder", "")

#         if hasattr(self.cfgs, "classes"):
#             if isinstance(self.cfgs.classes, list):
#                 setattr(self._vars, "classes", getattr(self.cfgs,  "classes"))
#             else:
#                 setattr(self._vars, "classes", str(getattr(self.cfgs,  "classes")))
#                 self. _vars.classes = list(self._vars.classes.split(','))
#         else:
#             setattr(self._vars, "classes", "")

#         print(">>>> vars:", self._vars)

#     def run(self, mode='train', img_exts=['bmp']):
#         imgs_info, num_data = get_images_info(mode=mode, img_folder=osp.join(self._vars.img_folder, mode), img_exts=img_exts, \
#                                 classes=self._vars.classes, roi_info=self._vars.roi_info, patch_info=self._vars.patch_info)

#         _info = []
#         num_rois_list, num_backgrounds_list = [], []
#         for img_info in imgs_info:
#             info = {}
#             info['img_file'] = img_info['img_file']
#             num_rois = len(img_info['rois'])
#             num_backgrounds = len(img_info['backgrounds'])

#             info['num_rois'] = num_rois
#             info['num_backgrounds'] = num_backgrounds

#             num_rois_list.append(num_rois)
#             num_backgrounds_list.append(num_backgrounds)

#             _info.append(info)

#         info = {'imgs_info': _info, 'total_num_imgs': len(_info), \
#                     'avg_num_rois': np.mean(num_rois_list), 'avg_num_backgrounds': np.mean(num_backgrounds_list)}

#         return info

# if __name__ == '__main__':
#     import os.path as osp
#     from pathlib import Path
#     FILE = Path(__file__).resolve()
#     ROOT = FILE.parents[0]

#     patch_info = PatchInfo()
#     recipe = osp.join(ROOT, '../data/pie.yaml')
#     print(recipe)
#     patch_info.alg_set_cfgs(recipe=recipe)
#     print(patch_info.cfgs)
#     data = patch_info.run()
#     print(data)



def save_df_by_class_to_pdf(performance_by_class, output_filename, idx2class=None):
    from reportlab.lib.pagesizes import letter
    from reportlab.lib.styles import getSampleStyleSheet
    from reportlab.platypus import SimpleDocTemplate, Paragraph
    from reportlab.lib import colors
    from reportlab.lib.units import inch

    # PDF 생성
    document = SimpleDocTemplate(output_filename, pagesize=letter)
    styles = getSampleStyleSheet()

    # 데이터 직접 PDF에 추가
    content = []

    # 제목 추가
    content.append(Paragraph('Performances', styles['Title']))

    # 데이터 추가
    for data in performance_by_class:
        for key, value in data.items():
            if isinstance(value, list):
                value_str = ', '.join(map(str, value))
            else:
                if key == 'class': 
                    value_str = str(idx2class[value])
                else:
                    value_str = str(value)
            content.append(Paragraph(f'{key}: {value_str}', styles['BodyText']))

        content.append(Paragraph('<br/><br/>', styles['Normal']))  # 추가적인 여백

    # PDF 생성
    document.build(content)


def save_pf_by_image_to_excel(performance_by_image, output_filename, idx2class=None):
    import pandas as pd
    from openpyxl import load_workbook
    from openpyxl.styles import PatternFill, Border, Side, Alignment
    
    # 데이터를 정리하여 리스트로 변환
    rows = []

    for filename, classes in performance_by_image.items():
        for class_id, metrics in classes.items():
            row = {'filename': filename, 'class': class_id if idx2class is None else idx2class[class_id]}
            row.update(metrics)
            rows.append(row)

    # 리스트를 DataFrame으로 변환
    df = pd.DataFrame(rows)

    # DataFrame을 엑셀 파일로 저장
    df.to_excel(output_filename, index=False, engine='openpyxl')


    # 엑셀 파일 로드
    workbook = load_workbook(output_filename)
    sheet = workbook.active

    # 현재 시트에서 'filename' 열 병합 수행
    filename_col = 'A'  # 'filename' 열이 A열에 있다고 가정

    # 병합할 시작 위치와 종료 위치 추적
    start_row = 2  # 데이터는 2행부터 시작하므로 2행부터 확인
    end_row = start_row

    fill_color = PatternFill(start_color="D9DDDC", end_color="D9DDDC", fill_type="solid")
    header_fill = PatternFill(start_color="D9DDDC", end_color="D9DDDC", fill_type="solid")  # 연한 파란색

    # 테두리 스타일 지정 (굵은 선)
    thick_border_rb = Border(right=Side(style='thick'), bottom=Side(style='thick'))
    thick_border_rt = Border(right=Side(style='thick'), top=Side(style='thick'))
    thick_border_r = Border(right=Side(style='thick'))
    thick_border_rtb = Border(right=Side(style='thick'), bottom=Side(style='thick'), top=Side(style='thick'))

    # 첫 번째 행(헤더)에 색상 적용
    for col in range(1, sheet.max_column + 1):
        cell = sheet.cell(row=1, column=col)
        cell.fill = header_fill
        cell.border = thick_border_rb

    # 행을 순회하면서 동일한 filename에 대해 병합 작업 수행
    for row in range(2, sheet.max_row + 1):
        sheet[f'A{row}'].fill = fill_color

        current_value = sheet[f'{filename_col}{row}'].value
        next_value = sheet[f'{filename_col}{row + 1}'].value if row + 1 <= sheet.max_row else None
        
        if current_value == next_value:
            end_row += 1
        else:
            if start_row != end_row:
                # 셀 병합
                sheet.merge_cells(f'{filename_col}{start_row}:{filename_col}{end_row}')
                
                # 병합된 범위에 대해 위아래 테두리 굵게 설정
                for merged_row in range(start_row, end_row + 1):
                    sheet[f'A{merged_row}'].border = thick_border_rb
                    
                for col in range(1, sheet.max_column + 1):
                    col_letter = sheet.cell(row=1, column=col).column_letter
                    for merged_row in range(start_row, end_row + 1):
                        if merged_row == start_row:
                            sheet[f'{col_letter}{merged_row}'].border = thick_border_rt
                        elif merged_row == end_row:
                            sheet[f'{col_letter}{merged_row}'].border = thick_border_rb
                        else:
                            sheet[f'{col_letter}{merged_row}'].border = thick_border_r
                    
            else:
                # 병합이 없는 경우에도 테두리와 색상 적용
                sheet[f'A{start_row}'].fill = fill_color
                sheet[f'A{start_row}'].border = thick_border_rtb
                
                for col in range(1, sheet.max_column + 1):
                    col_letter = sheet.cell(row=1, column=col).column_letter
                    sheet[f'{col_letter}{start_row}'].border = thick_border_rtb
                    
            start_row = row + 1
            end_row = start_row


    # 가운데 정렬 설정
    center_alignment = Alignment(horizontal='center', vertical='center')

    # 모든 셀에 대해 가운데 정렬 적용
    for row in sheet.iter_rows(min_row=1, max_row=sheet.max_row, min_col=1, max_col=sheet.max_column):
        for cell in row:
            cell.alignment = center_alignment


    # 엑셀 파일 저장
    workbook.save(output_filename)
    
    
if __name__ == '__main__':
    
    # 데이터 정의 (예제 데이터)
    performance_by_class = {
        'class': 'Class A',
        'accumulated_precision': [0.8, 0.85, 0.9],
        'accumulated_recall': [0.7, 0.75, 0.8],
        'precision': 0.9,
        'recall': 0.8,
        'ap': 0.85,
        'interpolated_precision': [0.8, 0.85, 0.9],
        'interpolated_recall': [0.7, 0.75, 0.8],
        'total_gt': 100,
        'total_tp': 80,
        'total_fp': 10,
        'total_fn': 20,
    }
    output_filename = '/HDD/datasets/projects/sungjin/body/benchmark/pf_by_class.pdf'
    
    save_df_by_class_to_pdf(performance_by_class, output_filename)


import glob 
import os
import os.path as osp
from ultralytics import YOLO

import imgviz
import json
import pandas as pd
from tqdm import tqdm
import numpy as np

from visionsuite.utils.visualizers.vis_obb import vis_obb


def test_obb(weights_file, imgsz, _classes, input_dir, output_dir, json_dir, compare_gt, iou_threshold, conf_threshold, line_width, font_scale,
                input_img_ext='bmp', output_img_ext='png', output_img_size_ratio=1):
    model = YOLO(weights_file) 

    _idx2class = {idx: cls for idx, cls in enumerate(_classes)}

    if not osp.exists(output_dir):
        os.makedirs(output_dir)

    img_files = glob.glob(osp.join(input_dir, f'*.{input_img_ext}'))

    preds = {}
    compare = {}
    for img_file in tqdm(img_files):
        filename = osp.split(osp.splitext(img_file)[0])[-1]
        pred = model(img_file, save=False, imgsz=imgsz, iou=iou_threshold, conf=conf_threshold, verbose=False)[0]
        
        idx2class = pred.names
        obb_result = pred.obb
        classes = obb_result.cls.tolist()
        confs = obb_result.conf.tolist()
        
        for res in obb_result:
            xywhr = res.xywhr
            angle = np.rad2deg(xywhr[0][-1].cpu().detach().item())
            # print("angle: ", angle)
            xyxyxyxy = res.xyxyxyxy
            # print("xyxyxyxyxy: ", xyxyxyxy)
        
        idx2xyxys = {}
        for cls, xyxys, conf in zip(classes, obb_result.xyxyxyxy, confs):
            if cls not in idx2xyxys:
                idx2xyxys[cls] = {'polygon': [], 'confidence': []}
            idx2xyxys[cls]['polygon'].append([xy.tolist() for xy in xyxys])
            idx2xyxys[cls]['confidence'].append(conf)
        
        if _classes is not None:
            new_idx2xyxys = {}
            for idx, _cls in enumerate(_classes):
                for jdx, cls in enumerate(idx2class.values()):
                    if cls == _cls and jdx in idx2xyxys:
                        new_idx2xyxys[idx] = idx2xyxys[jdx]
            
            idx2xyxys = new_idx2xyxys
            idx2class = _idx2class   
        
        color_map = imgviz.label_colormap()[1:len(idx2class) + 1 + 1]
        preds.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
        
        if compare_gt:
            _compare = vis_obb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
                               compare_gt=compare_gt, iou_threshold=iou_threshold, line_width=line_width, font_scale=font_scale,
                               output_img_ext=output_img_ext, output_img_size_ratio=output_img_size_ratio)
            _compare.update({"img_file": img_file})
            compare.update({filename: _compare})
        else:
            vis_obb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
                    compare_gt=compare_gt, iou_threshold=iou_threshold, line_width=line_width, font_scale=font_scale,
                    output_img_ext=output_img_ext, output_img_size_ratio=output_img_size_ratio)
                
    with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
        json.dump(preds, json_file, ensure_ascii=False, indent=4)

    if compare_gt:
        with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
            json.dump(compare, json_file, ensure_ascii=False, indent=4)
        
        df_compare = pd.DataFrame(compare)
        df_compare_pixel = df_compare.loc['diff_iou'].T
        df_compare_pixel.fillna(0, inplace=True)
        df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))



""" Full assembly of the parts to form the complete network """
import torch.nn.functional as F
from .unet_parts import *


class UNet(nn.Module):
    def __init__(self, n_channels, n_classes, bilinear=True):
        super(UNet, self).__init__()
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.bilinear = bilinear

        self.inc = DoubleConv(n_channels, 64)
        self.down1 = Down(64, 128)
        self.down2 = Down(128, 256)
        self.down3 = Down(256, 512)
        self.down4 = Down(512, 512)

        self.up1 = Up(1024, 256, bilinear)
        self.up2 = Up(512, 128, bilinear)
        self.up3 = Up(256, 64, bilinear)
        self.up4 = Up(128, 64, bilinear)
        self.outc = OutConv(64, n_classes)


    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        
        logits = self.outc(x)
        return logits



import glob 
import os
import os.path as osp
import torch
import imgviz
import json
import cv2
import numpy as np
import pandas as pd
from tqdm import tqdm

from mlearning.utils.vis.vis_hbb_patch import vis_hbb_patch
from mlearning.utils.functionals import letterbox
from athena.src.tasks.detection.frameworks.pytorch.models.yolov7.models.experimental import attempt_load as attempt_load7
from athena.src.tasks.detection.frameworks.pytorch.models.yolov5.utils.general import check_img_size
from athena.src.tasks.detection.frameworks.pytorch.models.yolov5.utils.general import (non_max_suppression, scale_coords)

compare_mask = True
imgsz = 1024
device = 'cuda'
weights = '/DeepLearning/_projects/sungjin_body/patch/2024_07_28_10_56_16/train/weights/best.pt'

input_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/w_json/기타'
json_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/w_json/기타'
output_dir = f'/DeepLearning/_projects/sungjin_body/tests/patch/winter/w_json/기타'

# input_dir = '/Data/01.Image/sungjin_yoke/IMAGE/BODY/24.07.29_미검이미지/wo_json/기타'
# json_dir = None
# output_dir = f'/DeepLearning/_projects/sungjin_body/tests/patch/winter/wo_json/기타'

patch_overlap_ratio = 0.1
patch_width = imgsz
patch_height = imgsz
dx = int((1. - patch_overlap_ratio) * patch_width)
dy = int((1. - patch_overlap_ratio) * patch_height)

compare_gt = True if json_dir is not None else False
model = attempt_load7(weights, map_location=device)  # load FP32 model
imgsz = check_img_size(imgsz, s=model.stride.max())  # check image size
if isinstance(imgsz, int):
    imgsz = (imgsz, imgsz)
model(torch.zeros(1, 3, imgsz[0], imgsz[1]).to(device).type_as(next(model.parameters())))  # run once
idx2class = {ii: vv for ii, vv in enumerate(model.classes)}

iou_threshold = 0.5
nms_conf_threshold = 0.25
nms_iou_threshold = 0.1
line_width = 5
font_scale = 2
classes = model.classes

if not osp.exists(output_dir):
    os.makedirs(output_dir)

img_files = glob.glob(osp.join(input_dir, '*.bmp'))

results = {}
compare = {}
for img_file in tqdm(img_files):
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    img = cv2.imread(img_file).astype(np.float32)
    img_h, img_w, img_c = img.shape
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)/255
    
    num_patches = 0
    idx2xyxys = {}
    for y0 in range(0, img_h, dy):
        for x0 in range(0, img_w, dx):
            num_patches += 1
            if y0 + patch_height > img_h:
                # skip if too much overlap (> 0.6)
                y = img_h - patch_height
            else:
                y = y0

            if x0 + patch_width > img_w:
                x = img_w - patch_width
            else:
                x = x0

            xmin, xmax, ymin, ymax = x, x + patch_width, y, y + patch_height
            input_img, ratio, pad_values = letterbox(img[ymin:ymax, xmin:xmax, :], (imgsz[0], imgsz[1]))
            
            torch_img = torch.from_numpy(np.transpose(input_img[np.newaxis, :], (0, 3, 1, 2))).to('cuda')
            with torch.no_grad():
                preds = model(torch_img)
            pred = preds[0].detach().cpu()
            dets = non_max_suppression(pred, nms_conf_threshold, nms_iou_threshold, None, False)
            
            for det in dets:
                if len(det):
                    det[:, :4] = scale_coords(img.shape[:2], det[:, :4], (img_h, img_w)).round()

                    for *xyxy, conf, cls in reversed(det):
                        cls = cls.detach().cpu().item()
                        conf = float(conf.detach().cpu().item())
                        if cls not in idx2xyxys.keys():
                            idx2xyxys[cls] = []
                            idx2xyxys[cls] = {'bbox': [], 'confidence': []}
                            
                        idx2xyxys[cls]['bbox'].append([[int(np.round(xyxy[0].detach().cpu().item()) + xmin), int(np.round(xyxy[1].detach().cpu().item()) + ymin)], 
                                            [int(np.round(xyxy[2].detach().cpu().item()) + xmin), int(np.round(xyxy[3].detach().cpu().item()) + ymin)]])
                        idx2xyxys[cls]['confidence'].append(conf)
        
    color_map = imgviz.label_colormap()[1:len(idx2class) + 1 + 1]
    results.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
    
    if compare_gt:
        _compare = vis_hbb_patch(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
                        compare_gt=compare_gt, iou_threshold=iou_threshold, line_width=line_width, font_scale=font_scale)
        _compare.update({"img_file": img_file})
        compare.update({filename: _compare})
    else:
        vis_hbb_patch(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, 
                compare_gt=compare_gt, iou_threshold=iou_threshold, line_width=line_width, font_scale=font_scale)
    
with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
    json.dump(results, json_file, ensure_ascii=False, indent=4)

if compare_gt:
    with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
        json.dump(compare, json_file, ensure_ascii=False, indent=4)
    
    df_compare = pd.DataFrame(compare)
    df_compare_pixel = df_compare.loc['diff_iou'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))


from .cifar_dataset import CifarDataset
from .load_data import load_data, image_folder #TODO: to be deprecated
from .directory_dataset import DirectoryDataset

__all__ = ['load_data', 'image_folder', 'DirectoryDataset', 'CifarDataset']


#!/usr/bin/env python3

from .builder import (
    DATASETS,
    PIPELINES,
    build_dataloader,
    build_dataset,
)
from .pipelines import *  # noqa: F401,F403

from .custom import (
    OTFCustomJointDataset,
    OTFCustomBinaryJointDataset,
)
from .cityscapes import OTFJointCityscapesDataset
from .mask_dataset import MaskDataset

__all__ = [
    "DATASETS",
    "PIPELINES",
    "build_dataloader",
    "build_dataset",
    "OTFCustomJointDataset",
    "OTFCustomBinaryJointDataset",
    "OTFJointCityscapesDataset",
    "MaskDataset",
]



import glob 
import os.path as osp 
import cv2 
import json
import os
from tqdm import tqdm 
import numpy as np
from copy import deepcopy
from shapely import Polygon, MultiPolygon

from visionsuite.utils.dataset.formats.labelme.utils import add_labelme_element, init_labelme_json, get_mask_from_labelme

def intersected_polygon(window, points):
    xmin, ymin = window[0]
    xmax, ymax = window[1]

    # 사각형 Window를 Polygon으로 정의
    window_polygon = Polygon([(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)])

    # Points를 Polygon으로 정의
    points_polygon = Polygon(points)

    # 교차 다각형 구하기 (intersection)
    if not window_polygon.is_valid:
        window_polygon = window_polygon.buffer(0)
    if not points_polygon.is_valid:
        points_polygon = points_polygon.buffer(0)
    geoms = window_polygon.intersection(points_polygon)
    
    if not geoms.is_empty:
        if geoms.geom_type == 'Polygon':
            return [[val1, val2] for val1, val2 in list(geoms.exterior.coords)]
        elif geoms.geom_type == 'Multipolygon':
            raise NotImplementedError
            return [list(x.exterior.coords) for x in geoms.geoms]
    else:
        return None


def intersection(boxA, boxB):
    # Box coordinates
    xmin1, ymin1 = boxA[0]
    xmax1, ymax1 = boxA[1]
    xmin2, ymin2 = boxB[0]
    xmax2, ymax2 = boxB[1]
    
    # Calculate the (x, y) coordinates of the intersection rectangle
    inter_xmin = max(xmin1, xmin2)
    inter_ymin = max(ymin1, ymin2)
    inter_xmax = min(xmax1, xmax2)
    inter_ymax = min(ymax1, ymax2)
    
    # Check if there is an intersection
    if inter_xmin < inter_xmax and inter_ymin < inter_ymax:
        return [[inter_xmin, inter_ymin], [inter_xmax, inter_ymax]]
    else:
        # No intersection
        return None


def min_max_normalize(image_array, min_val, max_val):
    normalized_array = (image_array - min_val) / (max_val - min_val)
    return np.clip(normalized_array, 0, 1)  # 값을 0과 1 사이로 클리핑


def labelme2patch(input_dir, output_dir, modes, rois, patch_width, patch_height, 
                    input_formats = ['bmp'],
                    output_format='bmp', patch_overlap_ratio = 0.2,
                    norm_val=None, vis=False, include_positive=True, classes_to_include=None):

    if not osp.exists(output_dir):
        os.makedirs(output_dir)
        
    if vis:
        vis_dir = osp.join(output_dir, 'vis')
        if not osp.exists(vis_dir):
            os.mkdir(vis_dir)
            
    dx = int((1. - patch_overlap_ratio) * patch_width)
    dy = int((1. - patch_overlap_ratio) * patch_height)
    class2label = {}
    for mode in modes:
        _output_dir = osp.join(output_dir, mode)
        if not osp.exists(_output_dir):
            os.mkdir(_output_dir)
        
        img_files = []
        for input_format in input_formats:
            img_files += glob.glob(osp.join(input_dir, mode, f'*.{input_format}'))
        for img_file in tqdm(img_files):
            filename = osp.split(osp.splitext(img_file)[0])[-1]
            json_file = osp.splitext(img_file)[0] + '.json'
            with open(json_file, 'r') as jf:
                anns = json.load(jf)['shapes']
            
            img = cv2.imread(img_file)
            img_h, img_w, img_c = img.shape
            
            if len(rois) == 0:
                rois = [[0, 0, img_w, img_h]]
            
            num_patches = 0
            for roi in rois:
                for y0 in range(roi[1], roi[3], dy):
                    for x0 in range(roi[0], roi[2], dx):
                        if y0 + patch_height > roi[3]:
                            y = roi[3] - patch_height
                        else:
                            y = y0

                        if x0 + patch_width > roi[2]:
                            x = roi[2] - patch_width
                        else:
                            x = x0
                
                        _labelme = init_labelme_json(filename + f'_{num_patches}.{output_format}', img_w, img_h)
                        xmin, xmax, ymin, ymax = x, x + patch_width, y, y + patch_height
                        window = [[xmin, ymin], [xmax, ymax]]

                        for ann in anns:
                            included = False
                            label = ann['label'].lower()
                            if classes_to_include:
                                if label not in classes_to_include:
                                    continue
                            
                            if label not in class2label:
                                class2label[label] = len(class2label) + 1
                            points = ann['points']
                            shape_type = ann['shape_type']
                                                        
                            if shape_type == 'point':
                                if not include_positive:
                                    continue
                                _points = []
                                for point in points:
                                    _points.append([point[0] - xmin, point[1] -ymin])
                                _labelme = add_labelme_element(_labelme, ann['shape_type'], ann['label'], _points)
                                included = True
                            
                            elif shape_type == 'rectangle':
                                intersected_points = intersection(window, points)
                                
                                if intersected_points:
                                    included = True 
                                    
                                    for intersected_point in intersected_points:
                                        intersected_point[0] -= xmin
                                        intersected_point[1] -= ymin
                                    
                                    _labelme = add_labelme_element(_labelme, ann['shape_type'], ann['label'], intersected_points)
                            elif shape_type == 'polygon':
                                if len(points) <= 2:
                                    if include_positive:
                                        _points = []
                                        for point in points:
                                            _points.append([point[0] - xmin, point[1] -ymin])
                                        _labelme = add_labelme_element(_labelme, ann['shape_type'], ann['label'], _points)
                                    continue
                                intersected_points = intersected_polygon(window, points)
                                if intersected_points:
                                    included = True 
                                    for intersected_point in intersected_points:
                                        intersected_point[0] -= xmin
                                        intersected_point[1] -= ymin
                                    _labelme = add_labelme_element(_labelme, ann['shape_type'], ann['label'], intersected_points)
                            
                                
                                
                        if included:
                            if norm_val is not None:
                                if norm_val['type'] == 'min_max':
                                    patch = min_max_normalize(deepcopy(img[ymin:ymax, xmin:xmax, :]), norm_val['min_val'], norm_val['max_val'])
                                    patch = (patch * 255).astype(np.uint8)
                            else:
                                patch = deepcopy(img[ymin:ymax, xmin:xmax, :])
                            cv2.imwrite(osp.join(_output_dir, filename + f'_{num_patches}.{output_format}'), patch)
                            with open(osp.join(_output_dir, filename + f'_{num_patches}.json'), 'w') as jf:
                                json.dump(_labelme, jf)
                                
                            if vis:
                                import imgviz
                                
                                mask = get_mask_from_labelme(osp.join(_output_dir, filename + f'_{num_patches}.json'),
                                                            patch_width, patch_height, class2label, format='opencv')
                                
                                vis_img = np.zeros((patch_height, patch_width*2, 3))
                                vis_img[:, :patch_width, :] = patch
                                color_map = imgviz.label_colormap(50)
                                mask = color_map[mask.astype(np.uint8)].astype(np.uint8)
                                vis_img[:, patch_width:, :] = mask 
                                
                                cv2.imwrite(osp.join(vis_dir, filename + f'_{num_patches}.png'), vis_img)
                                        
                            num_patches += 1


if __name__ == '__main__':
        
    input_dir = '/HDD/datasets/projects/Tenneco/Metalbearing/outer/250110/split_dataset'
    output_dir = '/HDD/datasets/projects/Tenneco/Metalbearing/outer/250110/split_patch_dataset'
    modes = ['train', 'val']
    classes_to_include = None


    patch_overlap_ratio = 0.2
    patch_width = 1120
    patch_height = 768
    rois = [[220, 60, 220 + 1120, 60 + 768]]
    vis = False
    include_positive = True
    input_formats = ['bmp']
    output_format = 'bmp'

    # norm_val = {'type': 'min_max', 'min_val': 44, 'max_val': 235}
    norm_val = None
        
    labelme2patch(input_dir, output_dir, modes, rois, patch_width, patch_height, input_formats, output_format=output_format,
                    norm_val=norm_val, vis=vis, include_positive=include_positive, classes_to_include=classes_to_include)
                            
                        
                    
                



                





import time 
import os.path as osp
from visionsuite.engines.utils.helpers import mkdir

def on_validator_build_start(validator, *args, **kwargs):
    pass

def on_validator_build_end(validator, *args, **kwargs):
    
    for attribute_name in validator.required_attributes:
        assert hasattr(validator, attribute_name), ValueError(f'{attribute_name} must be assgined in validator class')
        assert getattr(validator, attribute_name) is not None, ValueError(f"{attribute_name} is None for validator")
    

def on_validator_epoch_start(validator, *args, **kwargs):
    pass

def on_validator_epoch_end(validator, *args, **kwargs):
    # validator.metric_logger.synchronize_between_processes()

    def _save_val():
        if kwargs['epoch']%validator.archive.args['val']['save_freq_epoch'] == 0 and osp.exists(validator.archive.val_dir):
            vis_dir = osp.join(validator.archive.val_dir, str(kwargs['epoch']))
            if not osp.exists(vis_dir):
                mkdir(vis_dir)
               

            from visionsuite.engines.segmentation.utils.vis.vis_val import save_validation
            from visionsuite.engines.utils.functionals import denormalize

            save_validation(validator.model, validator.device, validator.dataloader.dataset, len(validator.label2index), 
                                kwargs['epoch'], vis_dir, denormalize=denormalize, input_channel=3, \
                            image_channel_order='bgr', validation_image_idxes_list=[])  

    def _save_results():
        validator.results.epoch = int(kwargs['epoch'])
    #     validator.results.loss = float(round(validator.metric_logger.meters['loss'].global_avg, 4))
    #     validator.results.accuracy = float(round(validator.metric_logger.meters["acc1"].global_avg, 4))
        validator.results.time_for_a_epoch = float(round(time.time() - kwargs['start_time_epoch'], 3))

    _save_val()
    _save_results()

def on_validator_batch_start(validator, *args, **kwargs):
    pass

def on_validator_batch_end(validator, *args, **kwargs):
    pass 

def on_validator_step_start(validator, *args, **kwargs): # iteration for a batch
    pass

def on_validator_step_end(validator, *args, **kwargs): # iteration for a batch
    pass

callbacks = {
    "on_validator_build_start": [on_validator_build_start],
    "on_validator_build_end": [on_validator_build_end],
    "on_validator_epoch_start": [on_validator_epoch_start],
    "on_validator_epoch_end": [on_validator_epoch_end],
    "on_validator_batch_start": [on_validator_batch_start],
    "on_validator_batch_end": [on_validator_batch_end],
    "on_validator_step_start": [on_validator_step_start],
    "on_validator_step_end": [on_validator_step_end],
}



########################################################
#                                                      #
#       author: omitted for anonymous submission       #
#                                                      #
#     credits and copyright coming upon publication    #
#                                                      #
########################################################


import argparse as ap
from collections import OrderedDict
import json
import os
import shutil

import numpy as np
from PIL import Image
from tqdm import tqdm
import cv2

from cityscapes import CityscapesBase


RGB_DIR = "leftImg8bit"
PARAMETERS_RAW_DIR = "camera"
DISPARITY_RAW_DIR = "disparity"
LABEL_DIR = "gtFine"


def save_indexed_png(filepath, label, colormap):
    # note that OpenCV is not able to handle indexed pngs correctly.
    img = Image.fromarray(np.asarray(label, dtype="uint8"))
    img.putpalette(list(np.asarray(colormap, dtype="uint8").flatten()))
    img.save(filepath, "PNG")


def get_files_by_extension(
    path, extension=".png", flat_structure=False, recursive=False, follow_links=True
):
    # check input args
    if not os.path.exists(path):
        raise IOError("No such file or directory: '{}'".format(path))

    if flat_structure:
        filelist = []
    else:
        filelist = {}

    # path is a file
    if os.path.isfile(path):
        basename = os.path.basename(path)
        if extension is None or basename.lower().endswith(extension):
            if flat_structure:
                filelist.append(path)
            else:
                filelist[os.path.dirname(path)] = [basename]
        return filelist

    # get filelist
    filter_func = lambda f: extension is None or f.lower().endswith(extension)
    for root, _, filenames in os.walk(path, topdown=True, followlinks=follow_links):
        filenames = list(filter(filter_func, filenames))
        if filenames:
            if flat_structure:
                filelist.extend((os.path.join(root, f) for f in filenames))
            else:
                filelist[root] = sorted(filenames)
        if not recursive:
            break

    # return
    if flat_structure:
        return sorted(filelist)
    else:
        return OrderedDict(sorted(filelist.items()))


if __name__ == "__main__":
    # argument parser
    parser = ap.ArgumentParser(
        description="Prepare Cityscapes dataset for segmentation."
    )
    parser.add_argument("output_path", type=str, help="path where to store dataset")
    parser.add_argument(
        "cityscapes_filepath",
        type=str,
        help="filepath to downloaded (and uncompressed) " "Cityscapes files",
    )
    args = parser.parse_args()

    # preprocess args and expand user
    output_path = os.path.expanduser(args.output_path)
    cityscapes_filepath = os.path.expanduser(args.cityscapes_filepath)

    # create output path if not exist
    os.makedirs(output_path, exist_ok=True)

    rgb_filepaths = get_files_by_extension(
        os.path.join(args.cityscapes_filepath, RGB_DIR),
        extension=".png",
        flat_structure=True,
        recursive=True,
    )

    label_filepaths = get_files_by_extension(
        os.path.join(args.cityscapes_filepath, LABEL_DIR),
        extension=".png",
        flat_structure=True,
        recursive=True,
    )
    label_filepaths = [
        fp for fp in label_filepaths if os.path.basename(fp).find("labelIds") > -1
    ]

    disparity_raw_filepaths = get_files_by_extension(
        os.path.join(args.cityscapes_filepath, DISPARITY_RAW_DIR),
        extension=".png",
        flat_structure=True,
        recursive=True,
    )

    parameters_filepaths = get_files_by_extension(
        os.path.join(args.cityscapes_filepath, PARAMETERS_RAW_DIR),
        extension=".json",
        flat_structure=True,
        recursive=True,
    )

    # check for consistency
    assert all(
        len(l) == 5000
        for l in [
            rgb_filepaths,
            label_filepaths,
            disparity_raw_filepaths,
            parameters_filepaths,
        ]
    )

    def get_basename(fp):
        # e.g. berlin_000000_000019_camera.json -> berlin_000000_000019
        return "_".join(os.path.basename(fp).split("_")[:3])

    basenames = [get_basename(f) for f in rgb_filepaths]
    for l in [label_filepaths, disparity_raw_filepaths, parameters_filepaths]:
        assert basenames == [get_basename(f) for f in l]

    filelists = {
        s: {
            "rgb": [],
            "disparity_raw": [],
            "labels_33": [],
            "labels_19": [],
        }
        for s in CityscapesBase.SPLITS
    }

    # copy rgb images
    print("Copying rgb files")
    for rgb_fp in tqdm(rgb_filepaths):
        basename = os.path.basename(rgb_fp)
        city = os.path.basename(os.path.dirname(rgb_fp))
        subset = os.path.basename(os.path.dirname(os.path.dirname(rgb_fp)))
        subset = "valid" if subset == "val" else subset

        dest_path = os.path.join(args.output_path, subset, CityscapesBase.RGB_DIR, city)
        os.makedirs(dest_path, exist_ok=True)

        # print(rgb_fp, '->', os.path.join(dest_path, basename))
        shutil.copy(rgb_fp, os.path.join(dest_path, basename))
        filelists[subset]["rgb"].append(os.path.join(city, basename))

    print("Processing label files")
    mapping_1plus33_to_1plus19 = np.array(
        [CityscapesBase.CLASS_MAPPING_REDUCED[i] for i in range(1 + 33)], dtype="uint8"
    )

    for l_fp in tqdm(label_filepaths):
        basename = os.path.basename(l_fp)
        city = os.path.basename(os.path.dirname(l_fp))
        subset = os.path.basename(os.path.dirname(os.path.dirname(l_fp)))
        subset = "valid" if subset == "val" else subset

        # load label with 1+33 classes
        label_full = cv2.imread(l_fp, cv2.IMREAD_UNCHANGED)

        # full: 1+33 classes (original label file -> just copy file)
        dest_path = os.path.join(
            args.output_path, subset, CityscapesBase.LABELS_FULL_DIR, city
        )
        os.makedirs(dest_path, exist_ok=True)
        # print(l_fp, '->', os.path.join(dest_path, basename))
        shutil.copy(l_fp, os.path.join(dest_path, basename))
        filelists[subset]["labels_33"].append(os.path.join(city, basename))

        # full: 1+33 classes colored
        dest_path = os.path.join(
            args.output_path, subset, CityscapesBase.LABELS_FULL_COLORED_DIR, city
        )
        os.makedirs(dest_path, exist_ok=True)
        save_indexed_png(
            os.path.join(dest_path, basename),
            label_full,
            colormap=CityscapesBase.CLASS_COLORS_FULL,
        )

        # map full to reduced: 1+33 classes -> 1+19 classes
        label_reduced = mapping_1plus33_to_1plus19[label_full]

        # reduced: 1+19 classes
        dest_path = os.path.join(
            args.output_path, subset, CityscapesBase.LABELS_REDUCED_DIR, city
        )
        os.makedirs(dest_path, exist_ok=True)
        cv2.imwrite(os.path.join(dest_path, basename), label_reduced)
        filelists[subset]["labels_19"].append(os.path.join(city, basename))

        # reduced: 1+19 classes colored
        dest_path = os.path.join(
            args.output_path, subset, CityscapesBase.LABELS_REDUCED_COLORED_DIR, city
        )
        os.makedirs(dest_path, exist_ok=True)
        save_indexed_png(
            os.path.join(dest_path, basename),
            label_reduced,
            colormap=CityscapesBase.CLASS_COLORS_REDUCED,
        )

    # ensure that filelists are valid and faultless
    def get_identifier(filepath):
        return "_".join(filepath.split("_")[:3])

    n_samples = 0
    for subset in CityscapesBase.SPLITS:
        identifier_lists = []
        for filelist in filelists[subset].values():
            identifier_lists.append([get_identifier(fp) for fp in filelist])

        assert all(l == identifier_lists[0] for l in identifier_lists[1:])
        n_samples += len(identifier_lists[0])

    assert n_samples == 5000

    # save meta files
    print("Writing meta files")
    np.savetxt(
        os.path.join(output_path, "class_names_1+33.txt"),
        CityscapesBase.CLASS_NAMES_FULL,
        delimiter=",",
        fmt="%s",
    )
    np.savetxt(
        os.path.join(output_path, "class_colors_1+33.txt"),
        CityscapesBase.CLASS_COLORS_FULL,
        delimiter=",",
        fmt="%s",
    )
    np.savetxt(
        os.path.join(output_path, "class_names_1+19.txt"),
        CityscapesBase.CLASS_NAMES_REDUCED,
        delimiter=",",
        fmt="%s",
    )
    np.savetxt(
        os.path.join(output_path, "class_colors_1+19.txt"),
        CityscapesBase.CLASS_COLORS_REDUCED,
        delimiter=",",
        fmt="%s",
    )

    for subset in CityscapesBase.SPLITS:
        subset_dict = filelists[subset]
        for key, filelist in subset_dict.items():
            np.savetxt(
                os.path.join(output_path, f"{subset}_{key}.txt"),
                filelist,
                delimiter=",",
                fmt="%s",
            )



# Copyright (c) OpenMMLab. All rights reserved.
from mmrotate.models.detectors.base import RotatedBaseDetector
from mmrotate.models.detectors.gliding_vertex import GlidingVertex
from mmrotate.models.detectors.oriented_rcnn import OrientedRCNN
from mmrotate.models.detectors.r3det import R3Det
from mmrotate.models.detectors.redet import ReDet
from mmrotate.models.detectors.roi_transformer import RoITransformer
from mmrotate.models.detectors.rotate_faster_rcnn import RotatedFasterRCNN
from mmrotate.models.detectors.rotated_fcos import RotatedFCOS
from mmrotate.models.detectors.rotated_reppoints import RotatedRepPoints
from mmrotate.models.detectors.rotated_retinanet import RotatedRetinaNet
from mmrotate.models.detectors.s2anet import S2ANet
from mmrotate.models.detectors.single_stage import RotatedSingleStageDetector
from mmrotate.models.detectors.two_stage import RotatedTwoStageDetector

from .rotated_imted import RotatedimTED

__all__ = [
    'RotatedRetinaNet', 'RotatedFasterRCNN', 'OrientedRCNN', 'RoITransformer',
    'GlidingVertex', 'ReDet', 'R3Det', 'S2ANet', 'RotatedRepPoints',
    'RotatedBaseDetector', 'RotatedTwoStageDetector',
    'RotatedSingleStageDetector', 'RotatedFCOS',

    'RotatedimTED'
]



import json
import os.path as osp
import random
from copy import deepcopy
from glob import glob
from threading import Thread

import numpy as np

from athena.src.data.slice.src.base_slicer import BaseSlicer
from athena.src.data.slice.src.make_patches import get_imgs_info_from_patches


class Slicer(BaseSlicer):
    """
    ### Attributes
        imgs_info: [
                        {
                            'img_file' : image file name, # string
                            'patches': [
                                        [x1, y1, x2, y2],
                                        [x1, y1, x2, y2],
                                        ...
                                    ], # list of lists: patch coordination
                            'counts': [0, 0, ...], # list
                            'backgrounds': [[x1, y1, x2, y2], [x1, y1, x2, y2], ...], # list of lists
                            'labels': [
                                        [{'label': 'abc', 'points': [[x1, y1], [x2, y2], ...], 'shape_type': '', 'roi': [x1, y1, x2, y2]},
                                        {'label': '', 'points': [[x1, y1], [x2, y2], ...], 'shape_type': '', 'roi': [x1, y1, x2, y2]}, ...],
                                        [{'label': '', 'points': [[x1, y1], [x2, y2], ...], 'shape_type': '', 'roi': [x1, y1, x2, y2]},
                                        {'label': '', 'points': [[x1, y1], [x2, y2], ...], 'shape_type': '', 'roi': [x1, y1, x2, y2]}, ...],
                                        ...
                                    ] # list of lists: label information in each patch coordination
                        },
                        {
                            'img_file' : image file name,
                            'patches': ...,
                            ...
                        },
                        ...
                    ]
            - Each patch coordinate is synched to each labels list and count.
            - Backgrounds exist only when using patch-based training
    """

    def __init__(
        self,
        mode,
        input_dir,
        img_exts=["bmp", "png", "jpg"],
        classes=None,
        roi_info=None,
        roi_from_json=False,
        patch_info=None,
        logger=None,
    ):
        super().__init__(__class__.__name__)

        self.mode = mode
        self.input_dir = input_dir
        self.img_exts = img_exts
        self.classes = classes
        self.roi_info = roi_info
        self.roi_from_json = roi_from_json
        self.patch_info = patch_info

        if logger is not None:
            self.set_log(
                log_dir=logger.log_dir,
                log_stream_level=logger.log_stream_level,
                log_file_level=logger.log_file_level,
            )

            self._logger.info(
                f"Logger for {__class__.__name__} has been set",
                self.__init__.__name__,
                __class__.__name__,
            )
            self._logger.info(
                f"****************************************************************************************"
            )
            self._logger.info(
                f"Slicer works on {mode} at {input_dir} within {img_exts} images for {classes}"
            )
            self._logger.info(f"RoI information: {roi_info}")
            self._logger.info(f"Patch information: {patch_info}")
            self._logger.info(
                f"****************************************************************************************"
            )
        else:
            self._logger = None

        self._img_files = []
        self._set_img_files()

        self.num_data = 0

    @property
    def num_img_files(self):
        return len(self._img_files)

    def run(self):
        for img_file in self._img_files:
            self.__img_info = {
                "img_file": img_file,
                "patches": [],
                "backgrounds": [],
                "labels": [],
            }
            self._set_roi_info(img_file)
            self._slice(img_file)

            if (
                self.__img_info["patches"] is not None
                and len(self.__img_info["patches"]) == 0
            ):
                print(f"*** There is zero patches for {img_file}")
                continue

            self.imgs_info = deepcopy(self.__img_info)

    def _slice(self, img_file):
        if self.patch_info is None:
            self._slice_wo_patch_info()
        else:
            self._slice_w_patch_info(img_file)

    def _slice_wo_patch_info(self):
        if self.roi_info is None:  # w/ roi & wo/ patch
            self.__img_info["patches"] = None
            self.__img_info["backgrounds"] = None
            self.__img_info["counts"] = [0]
            self.num_data += 1
        else:  # w/ roi & wo/ patch
            self.__img_info["backgrounds"] = None
            for roi in self.roi_info:
                self.__img_info["patches"].append(roi)
                self.num_data += 1
            self.__img_info["counts"] = [0] * len(self.__img_info["patches"])

    def _slice_w_patch_info(self, img_file):
        for roi in self.roi_info:
            patches, _num_data, background_patches, labels = get_imgs_info_from_patches(
                mode=self.mode,
                img_file=img_file,
                classes=self.classes,
                patch_info=self.patch_info,
                roi=roi,
                logger=self._logger,
            )
            self.__img_info["patches"] += patches
            self.__img_info["backgrounds"] += background_patches
            self.__img_info["labels"] += labels
            self.__img_info["counts"] = [0] * len(self.__img_info["patches"])
            self.num_data += _num_data

    def _set_img_files(self):
        if self._logger is not None:
            self._logger.assertion_log(
                self.mode is not None
                and self.input_dir is not None
                and self.img_exts != [],
                RuntimeError(
                    f"At least, one of mode({self.mode}), input-dir({self.input_dir}), and extension({self.img_exts}) is not defined"
                ),
                parent_class=__class__.__name__,
                parent_fn=self._set_img_files.__name__,
            )
        for ext in self.img_exts:
            if self.mode != "test":
                self._img_files += glob(
                    osp.join(self.input_dir, self.mode, "*.{}".format(ext))
                )
            else:
                self._img_files += glob(osp.join(self.input_dir, "*.{}".format(ext)))

        if self._logger is not None:
            self._logger.info(
                f"There are {len(self._img_files)} image files with {self.img_exts} in {self.mode} mode"
            )

    def _set_roi_info(self, img_file):
        with open(osp.splitext(img_file)[0] + ".json") as f:
            _json = json.load(f)
            if (
                self.roi_from_json
                and "rois" in _json.keys()
                and len(_json["rois"]) != 0
            ):
                self.roi_info = _json["rois"]
            elif self.roi_info is None:
                self.roi_info = [[]]

    def get_patch_information(self):
        _info = []
        num_patches_list, num_backgrounds_list = [], []
        for img_info in self.imgs_info:
            info = {}
            info["img_file"] = img_info["img_file"]
            num_patches = len(img_info["patches"])
            num_backgrounds = len(img_info["backgrounds"])

            info["num_patches"] = num_patches
            info["num_backgrounds"] = num_backgrounds

            num_patches_list.append(num_patches)
            num_backgrounds_list.append(num_backgrounds)

            _info.append(info)

        info = {
            "imgs_info": _info,
            "total_num_imgs": len(_info),
            "avg_num_patches_per_image": np.mean(num_patches_list),
            "avg_num_backgrounds_per_image": np.mean(num_backgrounds_list),
        }

        return info

    def save_imgs_info(self, output_dir):
        def save_as_json():
            with open(
                osp.join(output_dir, f"{self.mode}_imgs_info.json"), "w"
            ) as json_file:
                json.dump(self.imgs_info, json_file)

        Thread(target=save_as_json).start()

    def get_classes_info_by_patch(self, patch_ratio_by_class=None):

        def _labels_to_image_weights(class_counts, nc, class_weights):
            image_weights = (class_weights.reshape(1, nc) * class_counts).sum(1)
            # index = random.choices(range(n), weights=image_weights, k=1)  # weight image sample
            return image_weights

        class2idx = {
            _class.lower(): int(idx) + 1 for idx, _class in enumerate(self.classes)
        }
        patch_ratio_by_class = (
            np.ones(len(self.classes))
            if patch_ratio_by_class is None
            else patch_ratio_by_class
        )
        classes_info_by_patch = {
            "labels_in_patch": [],
            "labels_in_image": [],
            "positive_point": [],
            "cnt_class_by_patch": np.zeros(len(self.classes)),
        }
        num_background_patch_by_image, total_num_patch_by_image = 0, 0
        for idx, img_info in enumerate(self.imgs_info):
            num_background_patch_by_image += len(img_info["backgrounds"])
            total_num_patch_by_image += len(img_info["backgrounds"]) + len(
                img_info["patches"]
            )
            labels_in_image = [0] * len(self.classes)
            for patches, labels in zip(img_info["patches"], img_info["labels"]):
                labels_in_patch = [0] * len(self.classes)
                for label in labels:
                    if label["label"].lower() in class2idx.keys():
                        labels_in_image[class2idx[label["label"]] - 1] += 1
                    if len(label["points"]) == 0 and label["label"] == "":
                        classes_info_by_patch["positive_point"].append(idx)
                    else:
                        if label["label"] not in classes_info_by_patch:
                            classes_info_by_patch[label["label"]] = [idx]
                        else:
                            classes_info_by_patch[label["label"]].append(idx)
                classes_info_by_patch["labels_in_patch"].append(labels_in_patch)
            classes_info_by_patch["labels_in_image"].append(labels_in_image)

        _tmp_dict = {"total_num_instance": 0}
        for key, val in classes_info_by_patch.items():
            if key in class2idx.keys():
                _tmp_dict["num_" + key + "_instance"] = len(val)
                _tmp_dict["total_num_instance"] += len(val)
                _tmp_dict["num_" + key + "_image"] = len(set(val))
                _tmp_dict["max_" + key + "_instance"] = (
                    len(val) * patch_ratio_by_class[int(class2idx[key]) - 1]
                )
        _tmp_dict["num_positive_point_instance"] = len(
            classes_info_by_patch["positive_point"]
        )
        _tmp_dict["num_positive_point_image"] = len(
            set(classes_info_by_patch["positive_point"])
        )

        classes_info_by_patch.update(_tmp_dict)
        del _tmp_dict

        classes_info_by_patch["total_num_images"] = len(self.imgs_info)
        classes_info_by_patch["num_patches_per_image"] = int(
            total_num_patch_by_image / len(self.imgs_info)
        )
        classes_info_by_patch["avg_num_backgrounds_per_image"] = (
            num_background_patch_by_image / len(self.imgs_info)
        )

        # TODO
        # class-weights
        classes_info_by_patch["class_weights"] = [
            len(classes_info_by_patch[key])
            / classes_info_by_patch["total_num_instance"]
            for key in class2idx.keys()
        ]

        # image-weights
        classes_info_by_patch["image_weights"] = _labels_to_image_weights(
            classes_info_by_patch["labels_in_image"],
            len(self.classes),
            np.ones(len(self.classes))
            - np.array(classes_info_by_patch["class_weights"]),
        )
        classes_info_by_patch["image_indices"] = random.choices(
            range(len(self.imgs_info)),
            weights=classes_info_by_patch["image_weights"],
            k=len(self.imgs_info),
        )

        # patch-weights
        classes_info_by_patch["patch_weights"] = _labels_to_image_weights(
            classes_info_by_patch["labels_in_image"],
            len(self.classes),
            np.array(classes_info_by_patch["class_weights"]),
        )
        classes_info_by_patch["patch_indices"] = random.choices(
            range(len(self.imgs_info)),
            weights=classes_info_by_patch["patch_weights"],
            k=len(self.imgs_info),
        )

        return classes_info_by_patch

    def debug(self):
        pass



# from ..builder import DETECTORS
# from .two_stage import TwoStageDetector
from mmrotate.models.builder import ROTATED_DETECTORS
from mmrotate.models.detectors.two_stage import RotatedTwoStageDetector
import torch


@ROTATED_DETECTORS.register_module()
class RotatedimTED(RotatedTwoStageDetector):
    """Implementation of `imTED <https://arxiv.org/abs/2205.09613>`_"""

    def __init__(self,
                 backbone,
                 rpn_head,
                 roi_head,
                 train_cfg,
                 test_cfg,
                 roi_skip_fpn=False,
                 with_mfm=False,
                 neck=None,
                 pretrained=None,
                 proposals_dim=5):
        super(RotatedimTED, self).__init__(
            backbone=backbone,
            neck=neck,
            rpn_head=rpn_head,
            roi_head=roi_head,
            train_cfg=train_cfg,
            test_cfg=test_cfg,
            pretrained=pretrained)
        self.roi_skip_fpn = roi_skip_fpn
        self.with_mfm = with_mfm
        assert self.with_mfm == self.roi_head.with_mfm

        self.proposals_dim = proposals_dim

    def extract_feat(self, img):
        """Directly extract features from the backbone+neck."""
        x = self.backbone(img)
        if len(x) == 2:
            x, vit_feat = x
            if self.with_neck:
                x = self.neck(x)
            return x, vit_feat
        else:
            if self.with_neck:
                x = self.neck(x)
            return x

    def get_roi_feat(self, x, vit_feat):
        B, _, H, W = x[2].shape
        N = vit_feat.size(1)        # B, N, C
        M = N - H * W               # see whether there is a cls token or not
        assert M == 0 or M == 1     # check M
        if self.with_mfm:
            fea = [x[i] for i in range(len(x))]
            fea.append(vit_feat[:, M:, :].transpose(1, 2).reshape(B, -1, H, W).contiguous())
            x = tuple(fea)
        else:
            x = [
                vit_feat[:, M:, :].transpose(1, 2).reshape(B, -1, H, W).contiguous()
            ] if self.roi_skip_fpn else x
        return x

    def forward_train(self,
                      img,
                      img_metas,
                      gt_bboxes,
                      gt_labels,
                      gt_bboxes_ignore=None,
                      gt_masks=None,
                      proposals=None,
                      **kwargs):
        """
        Args:
            img (Tensor): of shape (N, C, H, W) encoding input images.
                Typically these should be mean centered and std scaled.

            img_metas (list[dict]): list of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                `mmdet/datasets/pipelines/formatting.py:Collect`.

            gt_bboxes (list[Tensor]): Ground truth bboxes for each image with
                shape (num_gts, 4) in [tl_x, tl_y, br_x, br_y] format.

            gt_labels (list[Tensor]): class indices corresponding to each box

            gt_bboxes_ignore (None | list[Tensor]): specify which bounding
                boxes can be ignored when computing the loss.

            gt_masks (None | Tensor) : true segmentation masks for each box
                used if the architecture supports a segmentation task.

            proposals : override rpn proposals with custom proposals. Use when
                `with_rpn` is False.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """
        x = self.extract_feat(img)
        if len(x) == 2:
            x, vit_feat = x[0], x[1]
            losses = dict()
            # RPN forward and loss
            if self.with_rpn:
                proposal_cfg = self.train_cfg.get('rpn_proposal',
                                                  self.test_cfg.rpn)
                rpn_losses, proposal_list = self.rpn_head.forward_train(
                    x,
                    img_metas,
                    gt_bboxes,
                    gt_labels=None,
                    gt_bboxes_ignore=gt_bboxes_ignore,
                    proposal_cfg=proposal_cfg)
                losses.update(rpn_losses)
            else:
                proposal_list = proposals

            roi_losses = self.roi_head.forward_train(self.get_roi_feat(x, vit_feat), img_metas, proposal_list,
                                                     gt_bboxes, gt_labels,
                                                     gt_bboxes_ignore, gt_masks,
                                                     **kwargs)
            losses.update(roi_losses)
            return losses
        else:
            losses = dict()
            # RPN forward and loss
            if self.with_rpn:
                proposal_cfg = self.train_cfg.get('rpn_proposal',
                                                  self.test_cfg.rpn)
                rpn_losses, proposal_list = self.rpn_head.forward_train(
                    x,
                    img_metas,
                    gt_bboxes,
                    gt_labels=None,
                    gt_bboxes_ignore=gt_bboxes_ignore,
                    proposal_cfg=proposal_cfg)
                losses.update(rpn_losses)
            else:
                proposal_list = proposals

            roi_losses = self.roi_head.forward_train(x, img_metas, proposal_list,
                                                     gt_bboxes, gt_labels,
                                                     gt_bboxes_ignore, gt_masks,
                                                     **kwargs)
            losses.update(roi_losses)
            return losses

    def forward_dummy(self, img):
        """Used for computing network flops.

        See `mmdetection/tools/analysis_tools/get_flops.py`
        """
        outs = ()
        # backbone
        x = self.extract_feat(img)
        if len(x) == 2:
            x, vit_feat = x[0], x[1]
        # rpn
        if self.with_rpn:
            rpn_outs = self.rpn_head(x)
            outs = outs + (rpn_outs, )
        # proposals = torch.randn(512, 4).to(img.device)
        proposals = torch.randn(1000, self.proposals_dim).to(img.device)
        # roi_head
        roi_outs = self.roi_head.forward_dummy(self.get_roi_feat(x, vit_feat), proposals)

        outs = outs + (roi_outs, )
        return outs
        
    async def async_simple_test(self,
                                img,
                                img_meta,
                                proposals=None,
                                rescale=False):
        """Async test without augmentation."""
        assert self.with_bbox, 'Bbox head must be implemented.'
        x = self.extract_feat(img)

        if proposals is None:
            proposal_list = await self.rpn_head.async_simple_test_rpn(
                x, img_meta)
        else:
            proposal_list = proposals

        return await self.roi_head.async_simple_test(
            x, proposal_list, img_meta, rescale=rescale)

    def simple_test(self, img, img_metas, proposals=None, rescale=False):
        """Test without augmentation."""
        assert self.with_bbox, 'Bbox head must be implemented.'

        x = self.extract_feat(img)
        if len(x) == 2:
            x, vit_feat = x[0], x[1]
            # get origin input shape to onnx dynamic input shape
            if torch.onnx.is_in_onnx_export():
                img_shape = torch._shape_as_tensor(img)[2:]
                img_metas[0]['img_shape_for_onnx'] = img_shape

            if proposals is None:
                proposal_list = self.rpn_head.simple_test_rpn(x, img_metas)
            else:
                proposal_list = proposals

            return self.roi_head.simple_test(
                self.get_roi_feat(x, vit_feat), proposal_list, img_metas, rescale=rescale)
        else:
            # get origin input shape to onnx dynamic input shape
            if torch.onnx.is_in_onnx_export():
                img_shape = torch._shape_as_tensor(img)[2:]
                img_metas[0]['img_shape_for_onnx'] = img_shape

            if proposals is None:
                proposal_list = self.rpn_head.simple_test_rpn(x, img_metas)
            else:
                proposal_list = proposals

            return self.roi_head.simple_test(
                x, proposal_list, img_metas, rescale=rescale)

    def aug_test(self, imgs, img_metas, rescale=False):
        """Test with augmentations.

        If rescale is False, then returned bboxes and masks will fit the scale
        of imgs[0].
        """
        raise NotImplementedError



import glob 
import os
import os.path as osp
import torch
import imgviz
import json
import cv2
import numpy as np
import pandas as pd
from tqdm import tqdm

from visionsuite.utils.visualizers.vis_obb import vis_obb
from athena.src.tasks.obbdetection.frameworks.pytorch.models.utils.nms import nms_rotated
from athena.src.tasks.obbdetection.frameworks.pytorch.models.utils.transforms import obb2poly_np_le90
from visionsuite.utils.helpers import letterbox, JsonEncoder


compare_mask = True
ml_framework = 'pytorch'
model_name='rtmdet'
backbone = 'large'
weights = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/outputs/outputs/OBBDETECTION/2024_10_11_16_05_36/train/weights/last.pt'
ckpt = torch.load(weights, map_location='cpu')
model = ckpt['model']
# model = de_parallel(model)
model.bbox_head.test_cfg = {
        'score_thr': 0.6,
        'iou_thr': 0.25,
        'max_per_img': 300,
        'agnostic_nms': False}

model.to('cuda:0')

compare_gt = True
iou_threshold = 0.7
width, height = 768, 768
max_dets = 50
nms_conf_threshold = 0.5
nms_iou_threshold = 0.25
classes = ['BOX']
idx2class = {idx: cls for idx, cls in enumerate(classes)}
_classes = ['BOX']
_idx2class = {idx: cls for idx, cls in enumerate(_classes)}
input_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/sfaw'
json_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/sfaw'
output_dir = f'/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/tests/sfaw/{model_name}_{backbone}'
# input_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/split_dataset_rich/val'
# json_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/split_dataset_rich/val'
# output_dir = f'/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/tests/rich/{model_name}_{backbone}'

if not osp.exists(output_dir):
    os.makedirs(output_dir)

img_files = glob.glob(osp.join(input_dir, '*.bmp'))

results = {}
compare = {}
for img_file in tqdm(img_files):
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    img = cv2.imread(img_file).astype(np.float32)
    img_h, img_w, img_c = img.shape
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)/255
    img, ratio, pad_values = letterbox(img, (width, height))
    
    torch_img = torch.from_numpy(np.transpose(img[np.newaxis, :], (0, 3, 1, 2))).to('cuda')
    model.forward = model.forward_export
    with torch.no_grad():
        preds = model(torch_img)
    pred = preds[0].detach().cpu()
    
    # start_angle = -90 / 180*np.pi
    # x, y, w, h, r = pred[:, :-1].unbind(-1)
    # _w = torch.where(w > h, w, h)
    # _h = torch.where(w > h, h, w)
    # r = torch.where(w > h, r, r + np.pi/2)
    # r = ((r - start_angle)%np.pi) + start_angle
    
    # pred[:, :-1] = torch.stack([x, y, _w, _h, r], dim=-1)
    
    dets, _ = nms_rotated(pred[:, :-1], pred[:, -1], iou_threshold=nms_iou_threshold)
    confs, labels = torch.max(dets[:, 5:], axis=1)

    idx2xyxys = {}
    for det, conf, cls in zip(dets[:max_dets].numpy(), confs[:max_dets].numpy(), labels[:max_dets].numpy()):
        if conf >= nms_conf_threshold:
            if cls not in idx2xyxys.keys():
                idx2xyxys[cls] = {'polygon': [], 'confidence': []}

            det[0] -= pad_values[0]
            det[1] -= pad_values[1]
            det[0] /= ratio[0]
            det[1] /= ratio[1]
            det[2] /= ratio[0]
            det[3] /= ratio[1]

            poly_points = det[np.newaxis, ...]
            poly_points = obb2poly_np_le90(poly_points)
            poly_points = poly_points[:, :8].reshape(4, 2)
            poly_points = np.round(poly_points, 2)

            idx2xyxys[cls]['polygon'].append([list(poly_point) for poly_point in poly_points])
            idx2xyxys[cls]['confidence'].append(conf)
  
    if _classes is not None:
        new_idx2xyxys = {}
        for idx, _cls in enumerate(_classes):
            for jdx, cls in enumerate(idx2class.values()):
                if cls == _cls:
                    new_idx2xyxys[idx] = idx2xyxys[jdx]
        
        idx2xyxys = new_idx2xyxys
        idx2class = _idx2class   
    
    color_map = imgviz.label_colormap()[1:len(idx2class) + 1 + 1]
    results.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
    
    if compare_gt:
        _compare = vis_obb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, compare_gt=compare_gt, iou_threshold=iou_threshold)
        _compare.update({"img_file": img_file})
        compare.update({filename: _compare})
    else:
        vis_obb(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir, compare_gt=compare_gt, iou_threshold=iou_threshold)
            
with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
    json.dump(results, json_file, ensure_ascii=False, indent=4, cls=JsonEncoder)

if compare_gt:
    with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
        json.dump(compare, json_file, ensure_ascii=False, indent=4)
    
    df_compare = pd.DataFrame(compare)
    df_compare_pixel = df_compare.loc['diff_iou'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))


#!/usr/bin/env python3

"""Implementation of DFF

https://github.com/Lavender105/DFF/blob/master/exps/models/dff.py
"""

import torch.nn as nn

from .base_multisupervision_head import BaseMultiSupervisionHead
from ..builder import HEADS
from ..utils import (
    GeneralizedLocationAdaptiveLearner,
    SideConv,
)


@HEADS.register_module()
class GeneralizedDFFHead(BaseMultiSupervisionHead):
    def __init__(
        self,
        edge_key="fuse",
        log_edge_keys=("fuse", "last"),
        binary_keys=[],
        multilabel_keys=("fuse", "last"),
        loss_binary=None,
        loss_multilabel=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            edge_key=edge_key,
            log_edge_keys=log_edge_keys,
            binary_keys=binary_keys,
            multilabel_keys=multilabel_keys,
            loss_binary=loss_binary,
            loss_multilabel=loss_multilabel,
            **kwargs,
        )

        _interp = "bilinear"  # nearest
        _bias = True

        # bias should not be turn on when some of the sides are not supervised

        sides = []

        for i in range(len(self.in_channels) - 1):
            sides.append(
                SideConv(
                    in_channels=self.in_channels[i],
                    out_channels=1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,  # bias instead of bn
                    bias=_bias,  # add bias in the last layer
                    act_cfg=self.act_cfg,
                    interpolation=_interp,
                    align_corners=self.align_corners,
                )
            )

        # last side is semantic
        sides.append(
            SideConv(
                in_channels=self.in_channels[-1],
                out_channels=self.num_classes,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,  # bias instead of bn
                bias=_bias,
                act_cfg=self.act_cfg,
                interpolation=_interp,
                align_corners=self.align_corners,
            )
        )

        self.sides = nn.ModuleList(sides)

        self.side_w = SideConv(
            in_channels=self.in_channels[-1],
            out_channels=self.num_classes * 4,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,  # might not need?
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )

        self.ada_learner = GeneralizedLocationAdaptiveLearner(
            num_sides=len(sides),
            in_channels=self.num_classes * 4,
            out_channels=self.num_classes * 4,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

    def forward(self, inputs):
        x = [i for i in inputs]
        assert isinstance(x, list)
        # get the input image size
        bs, c, h, w = x[-1].shape
        resize_to = (h, w)  # TODO: might be too large

        # remove the input image and unused features
        x = [x[i] for i in self.in_index]

        side_outs = []
        for i, layer in enumerate(self.sides):
            side_outs.append(layer(x[i], resize_to))

        last = side_outs[-1]

        side_outs.append(self.side_w(x[-1], resize_to))
        fuse = self.ada_learner(side_outs)

        return dict(fuse=fuse, last=last)


@HEADS.register_module()
class AuxDFFHead(GeneralizedDFFHead):
    ...



from visionsuite.engines.classification.utils.registry import OPTIMIZERS
from visionsuite.engines.utils.torch_utils.utils import set_weight_decay

    
def get_optim_parameters(model, **config):
    custom_keys_weight_decay = []
    if config['bias_weight_decay'] is not None:
        custom_keys_weight_decay.append(("bias", config['bias_weight_decay']))
    if config['transformer_embedding_decay'] is not None:
        for key in ["class_token", "position_embedding", "relative_position_bias_table"]:
            custom_keys_weight_decay.append((key, config['transformer_embedding_decay']))
            
    parameters = set_weight_decay(
        model,
        config['weight_decay'],
        norm_weight_decay=config['norm_weight_decay'],
        custom_keys_weight_decay=custom_keys_weight_decay if len(custom_keys_weight_decay) > 0 else None,
    )
    
    return parameters


def build_optimizer(model, **config):
    
    from visionsuite.engines.utils.helpers import get_params_from_obj
    optim_obj = OPTIMIZERS.get(config['type'])
    optim_params = get_params_from_obj(optim_obj)
    for key in optim_params.keys():
        if key in config:
            optim_params[key] = config[key]
            
        if key == 'params':
            optim_params[key] = get_optim_parameters(model, **config)
        
    return optim_obj(**optim_params)
    



import glob
import json
import os
import os.path as osp
import numpy as np
import cv2

input_dir = "/HDD/datasets/projects/sungjin/body/test/patches"
output_dir = "/HDD/datasets/projects/sungjin/body/test/vis"

if not osp.exists(output_dir):
    os.mkdir(output_dir)

# TODO: need to consider all shape-types

img_files = glob.glob(osp.join(input_dir, '*.bmp'))

for img_file in img_files:
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    json_file = osp.splitext(img_file)[0] + '.json'

    if osp.exists(json_file):
        img = cv2.imread(img_file)
        height, width, _ = img.shape

        with open(json_file) as jf:
            anns = json.load(jf)

        for ann in anns['shapes']:
            label = ann['label']
            shape_type = ann['shape_type']
            points = ann['points']

            if shape_type == 'rectangle':
                cv2.rectangle(img, (int(points[0][0]), int(points[0][1])), (int(points[1][0]), int(points[1][1])),
                            (0, 0, 255), 5)
            elif shape_type == 'polygon':
                cv2.fillPoly(img, [np.array(points, dtype=np.int32)], color=(255, 255, 0))
            else:
                raise NotImplementedError(f"There is no such shape-type({shape_type}) considered")


        # cv2.line(img, (0, 256), (width, 256), (0, 0, 255), 2) #horizontal
        # cv2.line(img, (408, 0), (408, height), (0, 0, 255), 2) #vertical
        # cv2.line(img, (664, 0), (664, height), (0, 0, 255), 2) #vertical
        # cv2.line(img, (512, 0), (512, height), (0, 0, 255), 2)

        cv2.imwrite(osp.join(output_dir, filename + '.bmp'), img)



import os.path as osp

import cv2
import numpy as np
import torch


def save_validation(
    model, dataloader, label2class, epoch, output_dir, device, preprocessor=None
):
    cnt = 0
    for batch in dataloader:
        if len(batch) == 3 or len(batch) == 4:
            for step_idx, (image, label, fname) in enumerate(
                zip(batch[0], batch[1], batch[2])
            ):
                label = str(label2class[np.argmax(np.array(label), axis=1)[0]])
                _image = np.expand_dims(image, axis=0)
                preds = model(_image)

                image = preprocessor.undo(image=image)["image"]
                # image = preprocessor.undo(image=image, key='normalize')['image']
                pred_label = str(label2class[np.argmax(np.array(preds), axis=1)[0]])

                height, width = image.shape[:2]
                if width >= 320 and width < 640:
                    font_scale = 0.3
                    thickness = 1
                elif width >= 640 and width < 1024:
                    font_scale = 0.6
                    thickness = 1
                else:
                    font_scale = 1
                    thickness = 2
                cv2.putText(
                    image,
                    "GT: " + str(label),
                    (2, 20),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    font_scale,
                    (255, 0, 0),
                    thickness,
                )
                cv2.putText(
                    image,
                    "PRED: " + str(pred_label),
                    (int(width / 2), 20),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    font_scale,
                    (255, 0, 255),
                    thickness,
                )

                if str(pred_label) == str(label):
                    cv2.putText(
                        image,
                        "OK",
                        (int(width / 2), int(height / 2)),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        font_scale * 3,
                        (0, 204, 0),
                        thickness * 3,
                    )
                else:
                    cv2.putText(
                        image,
                        "NG",
                        (int(width / 2), int(height / 2)),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        font_scale * 3,
                        (0, 0, 255),
                        thickness * 3,
                    )

                cv2.imwrite(
                    osp.join(
                        output_dir,
                        str(epoch) + "_" + fname + "_{}.png".format(step_idx),
                    ),
                    image,
                )
        elif len(batch) == 2:
            for step_idx, (image, label) in enumerate(zip(batch[0], batch[1])):
                # label = str(label2class[np.argmax(np.array(label), axis=1)[0]])
                label = str(label2class[label.item()])
                _image = torch.unsqueeze(image, axis=0).to(device)
                preds = model(_image)

                # image = preprocessor.undo(image, key="normalize")["image"]
                # image = preprocessor(np.transpose(image.numpy(), (1, 2, 0)))
                image = image.to('cpu')
                image = image.numpy()
                image = image.transpose((1, 2, 0))
                if preprocessor:
                    image = preprocessor(image)
                image = image.astype(np.uint8)
                image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
                # image = np.ascontiguousarray(image)
                pred_label = str(label2class[np.argmax(np.array(preds.cpu().detach().numpy()), axis=1)[0]])

                width = image.shape[1]
                if width < 640:
                    font_scale = 0.5
                    thickness = 1
                elif width >= 640 and width < 1024:
                    font_scale = 0.7
                    thickness = 1
                else:
                    font_scale = 1
                    thickness = 2

                cv2.putText(
                    image,
                    "GT: " + str(label),
                    (2, 20),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    font_scale,
                    (255, 0, 0),
                    thickness,
                )
                cv2.putText(
                    image,
                    "PRED: " + str(pred_label),
                    (int(width / 2), 20),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    font_scale,
                    (255, 0, 255),
                    thickness,
                )

                if str(pred_label) == str(label):
                    cv2.putText(
                        image,
                        "OK",
                        (width - 80, 50),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        font_scale*2,
                        (0, 204, 0),
                        thickness*2,
                    )
                else:
                    cv2.putText(
                        image,
                        "NG",
                        (width - 80, 50),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        font_scale*2,
                        (0, 0, 255),
                        thickness*2,
                    )

                cv2.imwrite(
                    osp.join(output_dir, str(epoch) + "_{}.png".format(cnt)), image
                )
                cnt += 1


import tensorflow as tf

class LabelSmoothSoftmaxCEV1(tf.keras.layers.Layer):
    def __init__(self, lb_smooth=0.1, reduction='mean', ignore_index=-100):
        super(LabelSmoothSoftmaxCEV1, self).__init__()
        self.lb_smooth = lb_smooth
        self.reduction = reduction
        self.lb_ignore = ignore_index

    # def call(self, logits, label):
    #     logits = tf.cast(logits, tf.float32)
    #     num_classes = tf.shape(logits)[1]
    #     label = tf.cast(label, tf.int32)
        
    #     # Reshape logits and label
    #     logits = tf.transpose(logits, [0, 2, 3, 1])  # (8, 384, 384, 19)
    #     label = tf.expand_dims(label, axis=-1)  # (8, 384, 384, 1)
        
    #     ignore = tf.equal(label, self.lb_ignore)
    #     n_valid = tf.reduce_sum(tf.cast(tf.logical_not(ignore), tf.float32))
        
    #     label = tf.where(ignore, tf.zeros_like(label), label)
    #     lb_pos, lb_neg = 1.0 - self.lb_smooth, self.lb_smooth / tf.cast(num_classes, tf.float32)
        
    #     lb_one_hot = tf.one_hot(tf.squeeze(label, axis=-1), depth=num_classes, on_value=lb_pos, off_value=lb_neg)
        
    #     logs = tf.nn.log_softmax(logits, axis=-1)
    #     loss = -tf.reduce_sum(logs * lb_one_hot, axis=-1)
    #     loss = tf.where(tf.squeeze(ignore, axis=-1), tf.zeros_like(loss), loss)
        
    #     if self.reduction == 'mean':
    #         loss = tf.reduce_sum(loss) / n_valid
    #     elif self.reduction == 'sum':
    #         loss = tf.reduce_sum(loss)
        
    #     return loss

    def call(self, logits, label):
        logits = tf.cast(logits, tf.float32)  # use fp32 to avoid nan
        
        num_classes = tf.shape(logits)[1]
        label = tf.identity(label)
        ignore = tf.equal(label, self.lb_ignore)
        n_valid = tf.reduce_sum(tf.cast(tf.logical_not(ignore), tf.float32))
        label = tf.where(ignore, tf.zeros_like(label), label)
        lb_pos, lb_neg = 1. - self.lb_smooth, self.lb_smooth / tf.cast(num_classes, tf.float32)

        lb_one_hot = tf.one_hot(label, depth=logits.shape[1], on_value=lb_pos, off_value=lb_neg)
        lb_one_hot = tf.reshape(lb_one_hot, tf.shape(logits))
        
        logs = tf.nn.log_softmax(logits, axis=1)
        loss = -tf.reduce_sum(logs * lb_one_hot, axis=1)
        loss = tf.where(ignore, tf.zeros_like(loss), loss)
        
        if self.reduction == 'mean':
            loss = tf.reduce_sum(loss) / n_valid
        elif self.reduction == 'sum':
            loss = tf.reduce_sum(loss)
        
        return loss

    
if __name__ == '__main__':
    
    import numpy as np
    
    np.random.seed(0)
    input_data = np.random.randn(8, 19, 384, 384).astype(np.float32)
    labels = np.random.randint(0, 19, size=(8, 384, 384)).astype(np.int64)
    
    tf_loss = LabelSmoothSoftmaxCEV1()
    tf_inputs = tf.constant(input_data)
    tf_labels = tf.constant(labels)
    tf_result = tf_loss(tf_inputs, tf_labels).numpy()
    print("tf: ", tf_result)
    
    import torch
    from visionsuite.utils.losses.torch.label_smooth import LabelSmoothSoftmaxCEV1 as TorchLabelSmoothSoftmaxCEV1
    
    torch_loss = TorchLabelSmoothSoftmaxCEV1()
    torch_inputs = torch.from_numpy(input_data)
    torch_labels = torch.from_numpy(labels).to(torch.int64)
    torch_result = torch_loss(torch_inputs, torch_labels)
    print("torch: ", torch_result)
    
    


from visionsuite.engines.utils.torch_utils.utils import collate_fn
from visionsuite.engines.segmentation.utils.registry import DATALOADERS
from visionsuite.engines.utils.helpers import get_params_from_obj

def build_dataloader(dataset, mode, **config):
    try:
        dataloader_obj = DATALOADERS.get(config['type'], case_sensitive=config['case_sensitive'])
        dataloader_params = get_params_from_obj(dataloader_obj)
        for key in dataloader_params.keys():
            if key == 'dataset':
                dataloader_params[key] = getattr(dataset, f'{mode}_{key}', dataset)
                
            if key in config:
                dataloader_params[key] = config[key]
            elif key in config[mode]:
                dataloader_params[key] = config[mode][key]
                
                
        dataloader = dataloader_obj(**dataloader_params)

    except Exception as error:
        raise Exception(f"{error} at build_dataloader")
    
    return dataloader




import argparse
from datetime import datetime
import json
import os
import sys
import numpy as np
from tqdm import tqdm

import torch
from torch.utils.tensorboard import SummaryWriter
import torch.optim
from src.args_test import ArgumentParser
from src.build_model import build_model
from src import utils
from src.prepare_data import prepare_data
import imgviz 
import cv2

from torchmetrics import JaccardIndex as IoU


def parse_args():
    parser = ArgumentParser(
        description="Open-World Semantic Segmentation (Training)",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.set_common_args()
    args = parser.parse_args()
    # The provided learning rate refers to the default batch size of 8.
    # When using different batch sizes we need to adjust the learning rate
    # accordingly:
    if args.batch_size != 8:
        args.lr = args.lr * args.batch_size / 8
        print(
            f"Notice: adapting learning rate to {args.lr} because provided "
            f"batch size differs from default batch size of 8."
        )

    return args


def train_main():
    args = parse_args()

    # directory for storing weights and other training related files
    training_starttime = datetime.now().strftime("%d_%m_%Y-%H_%M_%S-%f")
    ckpt_dir = os.path.join(
        args.results_dir, args.dataset, f"{args.id}", f"{training_starttime}"
    )
    os.makedirs(ckpt_dir, exist_ok=True)
    os.makedirs(os.path.join(ckpt_dir, "confusion_matrices"), exist_ok=True)

    with open(os.path.join(ckpt_dir, "args.json"), "w") as f:
        json.dump(vars(args), f, sort_keys=True, indent=4)

    with open(os.path.join(ckpt_dir, "argsv.txt"), "w") as f:
        f.write(" ".join(sys.argv))
        f.write("\n")

    # data preparation ---------------------------------------------------------
    data_loaders = prepare_data(args, ckpt_dir)

    train_loader, valid_loader, _ = data_loaders

    n_classes_without_void = train_loader.dataset.n_classes_without_void
    if args.class_weighting != "None":
        class_weighting = train_loader.dataset.compute_class_weights(
            weight_mode=args.class_weighting, c=args.c_for_logarithmic_weighting
        )
    else:
        class_weighting = np.ones(n_classes_without_void)
    # model building -----------------------------------------------------------
    model, device = build_model(args, n_classes=n_classes_without_void)

    # loss functions
    loss_function_train = utils.CrossEntropyLoss2d(
        weight=class_weighting, device=device
    )
    loss_objectosphere = utils.ObjectosphereLoss()
    loss_mav = utils.OWLoss(n_classes=n_classes_without_void)
    loss_contrastive = utils.ContrastiveLoss(n_classes=n_classes_without_void)

    pixel_sum_valid_data = valid_loader.dataset.compute_class_weights(
        weight_mode="linear"
    )
    pixel_sum_valid_data_weighted = np.sum(pixel_sum_valid_data * class_weighting)
    loss_function_valid = utils.CrossEntropyLoss2dForValidData(
        weight=class_weighting,
        weighted_pixel_sum=pixel_sum_valid_data_weighted,
        device=device,
    )

    train_loss = [loss_function_train, loss_objectosphere, loss_mav, loss_contrastive]
    val_loss = [loss_function_valid, loss_objectosphere, loss_mav, loss_contrastive]
    if not args.obj:
        train_loss[1] = None
        val_loss[1] = None
    if not args.mav:
        train_loss[2] = None
        val_loss[2] = None
    if not args.closs:
        train_loss[3] = None
        val_loss[3] = None
        
    if args.last_ckpt != "":
        ckpt = torch.load(args.last_ckpt)
        model.load_state_dict(ckpt['state_dict'])
        print("LOADED ckpt: ", args.last_ckpt)

    if args.load_weights != "":
        model.load_state_dict(torch.load(args.load_weights), strict=True)
        print("LOADED weights: ", args.load_weights)
        
    writer = SummaryWriter("runs/" + ckpt_dir.split(args.dataset)[-1])
    test(
        model=model,
        valid_loader=valid_loader,
        device=device,
        val_loss=val_loss,
        epoch=0,
        debug_mode=args.debug,
        writer=writer,
        classes=args.num_classes,
    )
    writer.flush()


def test(
    model,
    valid_loader,
    device,
    val_loss,
    epoch,
    writer,
    loss_function_valid_unweighted=None,
    debug_mode=False,
    classes=19,
):
    # set model to eval mode
    model.eval()

    # we want to store miou and ious for each camera
    miou = dict()
    ious = dict()

    loss_function_valid, loss_obj, loss_mav, loss_contrastive = val_loss

    # reset loss (of last validation) to zero
    loss_function_valid.reset_loss()

    if loss_function_valid_unweighted is not None:
        loss_function_valid_unweighted.reset_loss()

    compute_iou = IoU(
        task="multiclass", num_classes=classes, average="none", ignore_index=255
    ).to(device)

    mavs = None
    if loss_contrastive is not None:
        mavs = loss_mav.read()

    total_loss_obj = []
    total_loss_mav = []
    total_loss_con = []
    # validate each camera after another as all images of one camera have
    # the same resolution and can be resized together to the ground truth
    # segmentation size.
    color_map=imgviz.label_colormap(50)

    for i, sample in enumerate(tqdm(valid_loader, desc="Valid step")):
        # copy the data to gpu
        image = sample["image"].to(device)
        target = sample["label"].long().cuda() - 1
        target[target == -1] = 255

        if not device.type == "cpu":
            torch.cuda.synchronize()

        # forward pass
        with torch.no_grad():
            prediction_ss, prediction_ow = model(image)

            for (_image, _target, pred_ss) in zip(image, target, prediction_ss):
                vis_image = np.zeros((512, 1024*3, 3))
                
                vis_pred_ss = pred_ss.detach().cpu().numpy()
                vis_pred_ss = np.transpose(vis_pred_ss, (1, 2, 0))#.astype(np.uint8)
                vis_pred_ss = np.argmax(vis_pred_ss, axis=-1)
                vis_pred_ss = color_map[vis_pred_ss].astype(np.uint8)
                
                vis_image[:, :1024, :] = np.transpose(_image.detach().cpu().numpy(), (1, 2, 0))
                vis_image[:, 1024:1024*2, :] = color_map[_target.detach().cpu().numpy()].astype(np.uint8)
                vis_image[:, 1024*2:, :] = vis_pred_ss
                cv2.imwrite("/HDD/etc/contmav.png", vis_image)
                

            if not device.type == "cpu":
                torch.cuda.synchronize()

            compute_iou.update(prediction_ss, target.cuda())

            # compute valid loss
            loss_function_valid.add_loss_of_batch(
                prediction_ss, sample["label"].to(device)
            )

            loss_objectosphere = torch.tensor(0)
            loss_ows = torch.tensor(0)
            loss_con = torch.tensor(0)
            if loss_obj is not None:
                target_obj = sample["label"]
                target_obj[target_obj == 16] = 255
                target_obj[target_obj == 17] = 255
                target_obj[target_obj == 18] = 255
                loss_objectosphere = loss_obj(prediction_ow, sample["label"])
            total_loss_obj.append(loss_objectosphere.cpu().detach().numpy())
            if loss_mav is not None:
                loss_ows = loss_mav(prediction_ss, target.cuda(), is_train=False)
            total_loss_mav.append(loss_ows.cpu().detach().numpy())
            if loss_contrastive is not None:
                loss_con = loss_contrastive(mavs, prediction_ow, target, epoch)
            total_loss_con.append(loss_con.cpu().detach().numpy())

            if debug_mode:
                # only one batch while debugging
                break

    ious = compute_iou.compute().detach().cpu()
    miou = ious.mean()

    total_loss = (
        loss_function_valid.compute_whole_loss()
        + np.mean(total_loss_obj)
        + np.mean(total_loss_mav)
        + np.mean(total_loss_con)
    )
    print("LOSS: ", total_loss)
    print("Metrics/miou: ", miou)
    writer.add_scalar("Loss/val", total_loss, epoch)
    writer.add_scalar("Metrics/miou", miou, epoch)
    for i, iou in enumerate(ious):
        writer.add_scalar(
            "Class_metrics/iou_{}".format(i),
            torch.mean(iou),
            epoch,
        )

    return miou

    return 1 - best[0], best[1]


if __name__ == "__main__":
    train_main()




from visionsuite.engines.segmentation.utils.registry import LOOPS
from visionsuite.engines.segmentation.src.loops.train_loop import TrainLoop
from visionsuite.engines.utils.callbacks import Callbacks
from .callbacks import callbacks

@LOOPS.register()
class EpochBasedLoop(TrainLoop, Callbacks):
    def __init__(self, name="EpochBasedLoop"):
        TrainLoop.__init__(self, name=name)
        Callbacks.__init__(self)
        
        self.add_callbacks(callbacks)

    def build(self, _model, _dataset, _archive=None, *args, **kwargs):
        super().build(_model, _dataset, _archive=_archive, *args, **kwargs)
        
        self.run_callbacks('on_loop_build_end')
        
    def run(self):
        super().run()

        for epoch in range(self.start_epoch, self.epochs):
            self.set_epoch_for_sampler(epoch)
            for loop in self.loops:
                loop.run(epoch)
            
        self.run_callbacks('on_loop_run_end')